#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

__all__ = ("ScenarioTreeActionManagerPyro",)

import time
import itertools
import logging
from collections import defaultdict
import base64
try:
    import cPickle as pickle
except:
    import pickle

from pyomo.common.dependencies import attempt_import
from pyomo.opt.parallel.manager import ActionStatus
from pyomo.opt.parallel.pyro import PyroAsynchronousActionManager

pyu_pyro = attempt_import('pyutilib.pyro', alt_names=['pyu_pyro'])[0]
Pyro4 = attempt_import('Pyro4')[0]

import six
from six import advance_iterator, iteritems, itervalues

logger = logging.getLogger('pyomo.pysp')

#
# a specialized asynchronous action manager for the scenariotreeserver
#

class ScenarioTreeActionManagerPyro(PyroAsynchronousActionManager):

    def __init__(self, *args, **kwds):
        super(ScenarioTreeActionManagerPyro, self).__init__(*args, **kwds)
        # the SPPyroScenarioTreeServer objects associated with
        # this manager
        self.server_pool = []
        self._server_name_to_dispatcher_name = {}
        self._dispatcher_name_to_server_names = {}
        # tells the action manager to ignore task errors
        # (it will still report them, just take no action)
        self.ignore_task_errors = False

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """Close the manager."""
        if len(self.server_pool):
            self.release_servers()
        super(ScenarioTreeActionManagerPyro, self).close()

    def acquire_servers(self, servers_requested, timeout=None):

        if self._verbose:
            print("Attempting to acquire %s scenario tree servers"
                  % (servers_requested))
            if timeout is None:
                print("Timeout has been disabled")
            else:
                print("Automatic timeout in %s seconds" % (timeout))

        assert len(self.server_pool) == 0
        assert len(self._dispatcher_name_to_client) == 0
        assert len(self._server_name_to_dispatcher_name) == 0
        assert len(self._dispatcher_name_to_server_names) == 0
        assert len(self._dispatcher_proxies) == 0
        #
        # This process consists of the following steps:
        #
        # (1) Obtain the list of dispatchers from the nameserver
        # (2) Acquire all workers currently registered on each dispatcher
        # (3) Repeat (1) and (2) until we reach the timeout (if it exists)
        #     or until we obtain the number of servers requested
        # (4) Release any servers we don't need on dispatchers
        #
        wait_start = time.time()
        dispatcher_registered_servers = defaultdict(list)
        dispatcher_servers_to_release = defaultdict(list)
        dispatcher_proxies = {}
        servers_acquired = 0
        while servers_acquired < servers_requested:

            if (timeout is not None) and \
               ((time.time()-wait_start) > timeout):
                print("Timeout reached before %s servers could be acquired. "
                      "Proceeding with %s servers."
                      % (servers_requested, servers_acquired))
                break

            try:
                dispatchers = pyu_pyro.util.get_dispatchers(
                    host=self.host,
                    port=self.port,
                    caller_name="Client")
            except pyu_pyro.util._connection_problem:
                print("Failed to obtain one or more dispatchers from nameserver")
                continue
            for (name, uri) in dispatchers:
                dispatcher = None
                server_names = None
                if name not in dispatcher_proxies:
                    # connect to the dispatcher
                    if pyu_pyro.using_pyro3:
                        dispatcher = pyu_pyro.Pyro.core.getProxyForURI(uri)
                    else:
                        dispatcher = pyu_pyro.Pyro.Proxy(uri)
                        dispatcher._pyroTimeout = 10
                    try:
                        server_names = dispatcher.acquire_available_workers()
                    except pyu_pyro.util._connection_problem:
                        if pyu_pyro.using_pyro4:
                            dispatcher._pyroRelease()
                        else:
                            dispatcher._release()
                        continue
                    dispatcher_proxies[name] = dispatcher
                    if pyu_pyro.using_pyro4:
                        dispatcher._pyroTimeout = None
                else:
                    dispatcher = dispatcher_proxies[name]
                    server_names = dispatcher.acquire_available_workers()

                # collect the list of registered PySP workers
                servers_to_release = dispatcher_servers_to_release[name]
                registered_servers = dispatcher_registered_servers[name]
                for server_name in server_names:
                    if server_name.startswith("ScenarioTreeServerPyro_"):
                        registered_servers.append(server_name)
                    else:
                        servers_to_release.append(server_name)

                if (timeout is not None) and \
                   ((time.time()-wait_start) > timeout):
                    break

            servers_acquired = sum(len(_serverlist) for _serverlist
                                   in itervalues(dispatcher_registered_servers))
            # Don't overload the nameserver while trying to
            # collect dispatchers with registered workers.
            # If you haven't found them after the first few tries,
            # it's very likely that you are not going to.
            time.sleep(0.5)

        for name, servers_to_release in iteritems(dispatcher_servers_to_release):
            dispatcher_proxies[name].release_acquired_workers(servers_to_release)
        del dispatcher_servers_to_release

        #
        # Decide which servers we will utilize and do this in such a way
        # as to balance the workload we place on each dispatcher
        #
        server_to_dispatcher_map = {}
        dispatcher_servers_utilized = defaultdict(list)
        servers_utilized = 0
        dispatcher_names = itertools.cycle(dispatcher_registered_servers.keys())
        while servers_utilized < min(servers_requested, servers_acquired):
            name = advance_iterator(dispatcher_names)
            if len(dispatcher_registered_servers[name]) > 0:
                servername = dispatcher_registered_servers[name].pop()
                server_to_dispatcher_map[servername] = name
                dispatcher_servers_utilized[name].append(servername)
                servers_utilized += 1

        # copy the keys as we are modifying this list
        dispatcher_proxies_byURI = {}
        for name in list(dispatcher_proxies.keys()):
            dispatcher = dispatcher_proxies[name]
            servers = dispatcher_servers_utilized[name]
            if len(dispatcher_registered_servers[name]) > 0:
                # release any servers we do not need
                dispatcher.release_acquired_workers(
                    dispatcher_registered_servers[name])
            if len(servers) == 0:
                # release the proxy to this dispatcher,
                # we don't need it
                if pyu_pyro.using_pyro4:
                    dispatcher._pyroRelease()
                else:
                    dispatcher._release()
                del dispatcher_proxies[name]
            else:
                # when we initialize a client directly with a dispatcher
                # proxy it does not need to know the nameserver host or port
                client = self._create_client(dispatcher=dispatcher)
                self._dispatcher_name_to_server_names[client.URI] = servers
                dispatcher_proxies_byURI[client.URI] = dispatcher
                for servername in servers:
                    self._server_name_to_dispatcher_name[servername] = client.URI
                    self.server_pool.append(servername)
        self._dispatcher_proxies = dispatcher_proxies_byURI

    def release_servers(self):

        if self._verbose:
            print("Releasing scenario tree servers")

        for name in self._dispatcher_proxies:
            dispatcher = self._dispatcher_proxies[name]
            servers = self._dispatcher_name_to_server_names[name]
            # tell dispatcher that the servers we have acquired are no
            # longer needed
            dispatcher.release_acquired_workers(servers)

        self.server_pool = []
        self._server_name_to_dispatcher_name = {}
        self._dispatcher_name_to_server_names = {}

    #
    # Abstract Methods
    #

    def _get_dispatcher_name(self, queue_name):
        return self._server_name_to_dispatcher_name[queue_name]

    def _get_task_data(self, ah, **kwds):
        # Doing this serves two purposes:
        #   (1) It avoids issues with transmitting user-defined
        #       types over the wire that the dispatcher is not
        #       aware of (and therefore unable to de-serialize)
        #   (2) It improves performance on the dispatcher
        #       because de-serialization (and
        #       re-serialization) of raw bytes should be
        #       about as trivial as you can get for any
        #       serializer that Pyro/Pyro4 happens to be
        #       configured with (pickle is the fastest,
        #       but that is not the default in Pyro4 for
        #       security reasons).
        return pickle.dumps(kwds)

    def _download_results(self):

        found_results = False
        for client in itervalues(self._dispatcher_name_to_client):
            if len(self._dispatcher_name_to_client) == 1:
                # if there is a single dispatcher then we can do
                # a more efficient blocking call
                results = client.get_results(override_type=client.CLIENTNAME,
                                             block=True,
                                             timeout=None)
            else:
                results = client.get_results(override_type=client.CLIENTNAME,
                                             block=False)
            if len(results) > 0:
                found_results = True
                for task in results:
                    self.queued_action_counter -= 1

                    # The only reason we are go through this much
                    # effort to deal with the serpent serializer
                    # is because it is the default in Pyro4.
                    if pyu_pyro.using_pyro4 and \
                       (Pyro4.config.SERIALIZER == 'serpent'):
                        if six.PY3:
                            assert type(task['result']) is dict
                            assert task['result']['encoding'] == 'base64'
                            task['result'] = base64.b64decode(task['result']['data'])
                        else:
                            assert type(task['result']) is unicode
                            task['result'] = str(task['result'])
                    # ** See note in _get_task_data about why we pickle
                    #    all communication
                    task['result'] = pickle.loads(task['result'])

                    ah = self.event_handle.get(task['id'], None)
                    if ah is None:
                        # if we are here, this is really bad news!
                        raise RuntimeError(
                            "The %s found results for task with id=%s"
                            " - but no corresponding action handle "
                            "could be located! Showing task result "
                            "below:\n%s" % (type(self).__name__,
                                            task['id'],
                                            task.get('result', None)))
                    if type(task['result']) is pyu_pyro.TaskProcessingError:
                        ah.status = ActionStatus.error
                        self.event_handle[ah.id].update(ah)
                        msg = ("ScenarioTreeServer reported a processing "
                               "error for task with id=%s. Reason: \n%s"
                               % (task['id'], task['result'].args[0]))
                        if not self.ignore_task_errors:
                            raise RuntimeError(msg)
                        elif self.ignore_task_errors == 1:
                            logger.warning(msg)
                        # any value other than 0 or 1 will
                        # silently ignore task errors
                    else:
                        ah.status = ActionStatus.done
                        self.event_handle[ah.id].update(ah)
                        self.results[ah.id] = task['result']

        if not found_results:
            # If the queues are all empty, wait some time for things to
            # fill up. Constantly pinging dispatch servers wastes their
            # time, and inhibits task server communication. The good
            # thing about queues_to_check is that it simultaneously
            # grabs information for any queues with results => one
            # client query can yield many results.

            # TBD: We really need to parameterize the time-out value,
            #      but it isn't clear how to propagate this though the
            #      solver manager interface layers.
            time.sleep(0.01)



def pysp_postinit_callback(worker, scenario):
    print("postinit callback1: "+str(scenario.name)+", "
          +str(worker._aggregate_user_data['names']))
    print("")
    worker._aggregate_user_data['names'].remove(scenario.name)



"""Cut generation."""
from __future__ import division

from math import copysign

from pyomo.core import Constraint, minimize, value
from pyomo.core.expr import current as EXPR
from pyomo.contrib.gdpopt.util import copy_var_list_values, identify_variables
from pyomo.core.expr.taylor_series import taylor_series_expansion
from pyomo.core.expr import differentiate


def add_objective_linearization(solve_data, config):
    """Adds initial linearized objective in case it is nonlinear.

    This should be done for initializing the ECP method.

    """
    m = solve_data.working_model
    MindtPy = m.MindtPy_utils
    solve_data.mip_iter += 1
    gen = (obj for obj in MindtPy.jacs
           if obj is MindtPy.MindtPy_objective_expr)
    MindtPy.MindtPy_linear_cuts.mip_iters.add(solve_data.mip_iter)
    sign_adjust = 1 if MindtPy.obj.sense == minimize else -1
    # generate new constraints
    # TODO some kind of special handling if the dual is phenomenally small?
    for obj in gen:
        c = MindtPy.MindtPy_linear_cuts.ecp_cuts.add(
            expr=sign_adjust * sum(
                value(MindtPy.jacs[obj][id(var)]) * (var - value(var))
                for var in list(EXPR.identify_variables(obj.body))) +
            value(obj.body) <= 0)
        MindtPy.ECP_constr_map[obj, solve_data.mip_iter] = c


def add_oa_cuts(target_model, dual_values, solve_data, config,
                linearize_active=True,
                linearize_violated=True):
    """Linearizes nonlinear constraints.

    For nonconvex problems, turn on 'config.add_slack'. Slack variables will
    always be used for nonlinear equality constraints.
    """
    for (constr, dual_value) in zip(target_model.MindtPy_utils.constraint_list,
                                    dual_values):
        if constr.body.polynomial_degree() in (0, 1):
            continue

        constr_vars = list(identify_variables(constr.body))
        jacs = solve_data.jacobians

        # Equality constraint (makes the problem nonconvex)
        if constr.has_ub() and constr.has_lb() and constr.upper == constr.lower:
            sign_adjust = -1 if solve_data.objective_sense == minimize else 1
            rhs = constr.lower if constr.has_lb() and constr.has_ub() else rhs
            if config.add_slack:
                slack_var = target_model.MindtPy_utils.MindtPy_linear_cuts.slack_vars.add()
            target_model.MindtPy_utils.MindtPy_linear_cuts.oa_cuts.add(
                expr=copysign(1, sign_adjust * dual_value)
                * (sum(value(jacs[constr][var]) * (var - value(var))
                       for var in list(EXPR.identify_variables(constr.body)))
                    + value(constr.body) - rhs)
                - (slack_var if config.add_slack else 0) <= 0)

        else:  # Inequality constraint (possibly two-sided)
            if constr.has_ub() \
                and (linearize_active and abs(constr.uslack()) < config.zero_tolerance) \
                    or (linearize_violated and constr.uslack() < 0) \
                    or (config.linearize_inactive and constr.uslack() > 0):
                if config.add_slack:
                    slack_var = target_model.MindtPy_utils.MindtPy_linear_cuts.slack_vars.add()

                target_model.MindtPy_utils.MindtPy_linear_cuts.oa_cuts.add(
                    expr=(sum(value(jacs[constr][var])*(var - var.value)
                              for var in constr_vars) + value(constr.body)
                          - (slack_var if config.add_slack else 0)
                          <= constr.upper)
                )

            if constr.has_lb() \
                and (linearize_active and abs(constr.lslack()) < config.zero_tolerance) \
                    or (linearize_violated and constr.lslack() < 0) \
                    or (config.linearize_inactive and constr.lslack() > 0):
                if config.add_slack:
                    slack_var = target_model.MindtPy_utils.MindtPy_linear_cuts.slack_vars.add()

                target_model.MindtPy_utils.MindtPy_linear_cuts.oa_cuts.add(
                    expr=(sum(value(jacs[constr][var])*(var - var.value)
                              for var in constr_vars) + value(constr.body)
                          + (slack_var if config.add_slack else 0)
                          >= constr.lower)
                )

# def add_oa_equality_relaxation(var_values, duals, solve_data, config, ignore_integrality=False):
#     """More general case for outer approximation

#     This method covers nonlinear inequalities g(x)<=b and g(x)>=b as well as
#     equalities g(x)=b all in the same linearization call. It combines the dual
#     with the objective sense to figure out how to generate the cut.
#     Note that the dual sign is defined as follows (according to IPOPT):
#       sgn  | min | max
#     -------|-----|-----
#     g(x)<=b|  +1 | -1
#     g(x)>=b|  -1 | +1

#     Note additionally that the dual value is not strictly neccesary for inequality
#     constraints, but definitely neccesary for equality constraints. For equality
#     constraints the cut will always be generated so that the side with the worse objective
#     function is the 'interior'.

#     ignore_integrality: Accepts float values for discrete variables.
#                         Useful for cut in initial relaxation
#     """

#     m = solve_data.mip
#     MindtPy = m.MindtPy_utils
#     MindtPy.MindtPy_linear_cuts.nlp_iters.add(solve_data.nlp_iter)
#     sign_adjust = -1 if solve_data.objective_sense == minimize else 1

#     copy_var_list_values(from_list=var_values,
#                          to_list=MindtPy.variable_list,
#                          config=config,
#                          ignore_integrality=ignore_integrality)

#     # generate new constraints
#     # TODO some kind of special handling if the dual is phenomenally small?
#     # TODO-romeo conditional for 'global' option, i.e. slack or no slack
#     jacs = solve_data.jacobians
#     for constr, dual_value in zip(MindtPy.constraint_list, duals):
#         if constr.body.polynomial_degree() in (1, 0):
#             continue
#         rhs = ((0 if constr.upper is None else constr.upper)
#                + (0 if constr.lower is None else constr.lower))
#         # Properly handle equality constraints and ranged inequalities
#         # TODO special handling for ranged inequalities? a <= x <= b
#         rhs = constr.lower if constr.has_lb() and constr.has_ub() else rhs
#         slack_var = MindtPy.MindtPy_linear_cuts.slack_vars.add()
#         MindtPy.MindtPy_linear_cuts.oa_cuts.add(
#             expr=copysign(1, sign_adjust * dual_value)
#             * (sum(value(jacs[constr][var]) * (var - value(var))
#                    for var in list(EXPR.identify_variables(constr.body)))
#                + value(constr.body) - rhs)
#             - slack_var <= 0)


def add_int_cut(var_values, solve_data, config, feasible=False):
    if not config.add_integer_cuts:
        return

    config.logger.info("Adding integer cuts")

    m = solve_data.mip
    MindtPy = m.MindtPy_utils
    int_tol = config.integer_tolerance

    binary_vars = [v for v in MindtPy.variable_list if v.is_binary()]

    # copy variable values over
    for var, val in zip(MindtPy.variable_list, var_values):
        if not var.is_binary():
            continue
        var.value = val

    # check to make sure that binary variables are all 0 or 1
    for v in binary_vars:
        if value(abs(v - 1)) > int_tol and value(abs(v)) > int_tol:
            raise ValueError('Binary {} = {} is not 0 or 1'.format(
                v.name, value(v)))

    if not binary_vars:  # if no binary variables, skip.
        return

    int_cut = (sum(1 - v for v in binary_vars
                   if value(abs(v - 1)) <= int_tol) +
               sum(v for v in binary_vars
                   if value(abs(v)) <= int_tol) >= 1)

    MindtPy.MindtPy_linear_cuts.integer_cuts.add(expr=int_cut)

    # TODO need to handle theoretical implications of backtracking
    # if not feasible:
    #     # Add the integer cut
    #     MindtPy.MindtPy_linear_cuts.integer_cuts.add(expr=int_cut)
    # else:
    #     MindtPy.MindtPy_linear_cuts.feasible_integer_cuts.add(expr=int_cut)



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

from __future__ import division

import os
import logging
import copy

from six import iteritems
from operator import itemgetter

from pyomo.pysp.phutils import *
from pyomo.opt import UndefinedData

logger = logging.getLogger('pyomo.pysp')

# ===== various ways to extract xhat =======
def ExtractInternalNodeSolutionsWithClosestScenarioNodebyNode(ph):
    # find the scenario closest to xbar at each node and return it

    node_solutions = {}

    def ScenXbarDist(scen, tree_node):
        # crude estimate of stdev to get a approx, truncated z score

        dist = 0
        xbars = tree_node._xbars
        mins = tree_node._minimums
        maxs = tree_node._maximums
        for variable_id in tree_node._standard_variable_ids:
            diff = scen._x[tree_node._name][variable_id] - xbars[variable_id]
            s_est = (maxs[variable_id] - mins[variable_id]) / 4.0 # close enough to stdev
            if s_est > ph._integer_tolerance:
                dist += min(3, abs(diff)/s_est) # truncated z score
        return dist

    ClosestScen = None
    ClosestScenDist = None
    for stage in ph._scenario_tree._stages[:-1]:
        for tree_node in stage._tree_nodes:
            this_node_sol = node_solutions[tree_node._name] = {}
            for scenario in tree_node._scenarios:
                if ClosestScenDist == 0:
                    break
                thisdist = ScenXbarDist(scenario, tree_node)
                if ClosestScenDist == None or thisdist < ClostScenDist:
                     ClosestScendist = thisdist
                     ClosestScen = scenario

            for variable_id in tree_node._standard_variable_ids:
                ## print ("extracting for "+str(variable_id)+" the value "+str(ClosestScen._x[tree_node._name][variable_id]))
                this_node_sol[variable_id] = ClosestScen._x[tree_node._name][variable_id]

    return node_solutions

def ExtractInternalNodeSolutionsWithDiscreteRounding(ph):

    node_solutions = {}
    for stage in ph._scenario_tree._stages[:-1]:
        for tree_node in stage._tree_nodes:
            this_node_sol = node_solutions[tree_node._name] = {}
            xbars = tree_node._xbars
            for variable_id in tree_node._standard_variable_ids:
                if not tree_node.is_variable_discrete(variable_id):
                    this_node_sol[variable_id] = xbars[variable_id]
                else:
                    # rounded xbar, which has a MUCH
                    # better chance of being feasible
                    this_node_sol[variable_id] = \
                        int(round(xbars[variable_id]))

    return node_solutions

def ExtractInternalNodeSolutionsWithDiscreteVoting(ph):

    node_solutions = {}
    for stage in ph._scenario_tree._stages[:-1]:
        for tree_node in stage._tree_nodes:
            this_node_sol = node_solutions[tree_node._name] = {}
            xbars = tree_node._xbars
            for variable_id in tree_node._standard_variable_ids:
                if not tree_node.is_variable_discrete(variable_id):
                    this_node_sol[variable_id] = xbars[variable_id]
                else:
                    # for discrete variables use a weighted vote
                    # Note: for binary this can just be computed
                    #       by rounding the xbar (assuming it's an
                    #       average).  However, the following
                    #       works for binary and general integer,
                    #       where rounding the average is not
                    #       necessarily the same as a weighted vote
                    #       outcome.
                    vals = [int(round(scenario._x[tree_node._name][variable_id]))\
                            for scenario in tree_node._scenarios]
                    bins = list(set(vals))
                    vote = []
                    for val in bins:
                        vote.append(sum(scenario._probability \
                                        for scenario in tree_node._scenarios \
                                        if int(round(scenario._x[tree_node._name][variable_id])) == val))

                    # assign the vote outcome
                    this_node_sol[variable_id] = bins[vote.index(max(vote))]

    return node_solutions

def ExtractInternalNodeSolutionsWithSlamming(ph):
    from pyomo.pysp.plugins.wwphextension import _parse_yaml_file
    # Since it was a file, 
    #   assume that the argument was a json file with slamming instructions.
    # This will ignore suffixes we don't care about.
    # If there are no instructions use xbar.
    # Note: there is an implicit pecking order.
    print ("For x-hat, using slamming suffixes in",ph._xhat_method)
    slamdict = {}
    for suffix_name, suffix_value, variable_ids in \
          _parse_yaml_file(ph, ph._xhat_method):
        for node_name, node_variable_ids in iteritems(variable_ids):
            for variable_id in node_variable_ids:
                if variable_id not in slamdict:
                    slamdict[variable_id] = {}
                slamdict[variable_id][suffix_name] = suffix_value

    verbose = ph._verbose
    node_solutions = {}
    for stage in ph._scenario_tree._stages[:-1]:
        for tree_node in stage._tree_nodes:
            this_node_sol = node_solutions[tree_node._name] = {}
            xbars = tree_node._xbars
            mins = tree_node._minimums
            maxs = tree_node._maximums
            warnb = False  # did the user do something less than cool?

            for variable_id in tree_node._standard_variable_ids:
                if verbose:
                    variable_name, index = tree_node._variable_ids[variable_id]
                    full_variable_name = variable_name+indexToString(index)
                    print ("Setting x-hat for",full_variable_name)
                if variable_id not in slamdict or slamdict[variable_id]['CanSlamToAnywhere']:
                    if not tree_node.is_variable_discrete(variable_id):
                        this_node_sol[variable_id] = xbars[variable_id]
                        if verbose:
                            print ("   x-bar", this_node_sol[variable_id]) 
                    else:
                        this_node_sol[variable_id] = int(round(xbars[variable_id]))
                        if verbose:
                            print ("   rounded x-bar", this_node_sol[variable_id]) 
                elif slamdict[variable_id]['CanSlamToMin']:
                    this_node_sol[variable_id] = mins[variable_id]
                    if verbose:
                        print ("   min over scenarios", this_node_sol[variable_id]) 
                elif slamdict[variable_id]['CanSlamToMax']:
                    this_node_sol[variable_id] = maxs[variable_id]
                    if verbose:
                        print ("   max over scenarios", this_node_sol[variable_id]) 
                elif slamdict[variable_id]['CanSlamToLB']:
                    warnb = True
                    this_node_sol[variable_id] = mins[variable_id]
                    if verbose:
                        print ("   Lower Bound", this_node_sol[variable_id]) 
                elif slamdict[variable_id]['CanSlamToUB']:
                    warnb = True
                    this_node_sol[variable_id] = maxs[variable_id]
                    if verbose:
                        print ("   Upper Bound", this_node_sol[variable_id]) 
    if warnb:
        print ("Warning: for xhat determination from file %s, some variables had an upper or lower bound slam but not a corresponding min or max", ph._xhat_method)
    return node_solutions


#### call one ####

def ExtractInternalNodeSolutionsforInner(ph):
    # get xhat for the inner bound (upper bound for a min problem)
    if ph._xhat_method == "closest-scenario":
        return ExtractInternalNodeSolutionsWithClosestScenarioNodebyNode(ph)
    elif ph._xhat_method == "voting":
        return ExtractInternalNodeSolutionsWithDiscreteVoting(ph)
    elif ph._xhat_method == "rounding":
        return ExtractInternalNodeSolutionsWithDiscreteRounding(ph)
    elif os.path.isfile(ph._xhat_method):
        return ExtractInternalNodeSolutionsWithSlamming(ph)
    else:
        raise RuntimeError("Unknown x-hat determination method=%s specified in PH - unable to extract candidate solution for inner bound computation" % ph._xhat_method)

#============== end xhat code ==========

class _PHBoundBase(object):

    # Nothing interesting
    STATUS_NONE                      = 0b0000
    # Used mipgap
    STATUS_MIPGAP                    = 0b0001
    # Solution gap was not reported
    STATUS_GAP_NA                    = 0b0010
    # Solution has nonzero
    # optimality gap
    STATUS_GAP_NONZERO               = 0b0100
    # One or more subproblems were infeasible
    STATUS_SOLVE_FAILED              = 0b1000

    WARNING_MESSAGE = {}

    # No mipgap detected, but a
    # nonzero solution gap was
    # found
    WARNING_MESSAGE[0b0100] = \
        "** Possibly Conservative - Mipgap Unknown, And Nonzero Solution Gap Reported **"

    # Used mipgap and solver did
    # report a solution gap
    WARNING_MESSAGE[0b0101] = \
        "** Possibly Conservative - Mipgap Detected, And Nonzero Solution Gap Reported **"

    # Used mipgap and solver did NOT
    # report a solution gap
    WARNING_MESSAGE[0b0111] = \
        "** Extreme Caution - Mipgap Detected, But No Solution Gap Information Obtained - Bound May Be Invalid **"
    WARNING_MESSAGE[0b0011] = \
        "** Extreme Caution - Mipgap Detected, But No Solution Gap Information Obtained - Bound May Be Invalid **"

    WARNING_MESSAGE[0b0110] = \
        "** Caution - Solver Did Not Provide Solution Gap Information - Bound May Be Invalid **"
    WARNING_MESSAGE[0b0010] = \
        "** Caution - Solver Did Not Provide Solution Gap Information - Bound May Be Invalid **"

    # Tags for operations for which we need to undo in order to return
    # ph to its original state
    SOLUTION_CACHING     = (1,)
    VARIABLE_FREEING     = (2,)
    DEACTIVATE_PROXIMAL  = (3,)
    DEACTIVATE_WEIGHT    = (4,)
    CACHE_WEIGHTS        = (5,)
    TREE_VARIABLE_FIXING = (6,)

    def __init__(self):

        # the interval for which bound computations are performed
        # during the ph iteration k solves.  A bound update is always
        # performed at iteration 0 with the (non-PH-augmented)
        # objective and the update interval starts with a bound update
        # at ph iteration 1.
        self._update_interval = 1

        # keys are ph iteration except for the trival bound whose key
        # is None
        self._outer_bound_history = {}
        self._outer_status_history = {}
        self._inner_bound_history = {}
        self._inner_status_history = {}

        self._is_minimizing = True

        # stack of operations modifying ph
        self._stack = []

    def RestoreLastPHChange(self, ph):
        if len(self._stack):
            tmp_restore_stack = [self._stack.pop()]
            real_stack = self._stack
            self._stack = tmp_restore_stack
            self.RestorePH(ph)
            self._stack = real_stack

    def RestorePH(self, ph):

        while self._stack:

            op, op_data = self._stack.pop()

            if op == self.SOLUTION_CACHING:

                ph.restoreCachedSolutions(op_data, release_cache=True)

            elif op == self.VARIABLE_FREEING:

                ph_fixed, ph_fix_queue = op_data

                # Refix all previously fixed variables
                for tree_node in ph._scenario_tree._tree_nodes:
                    for variable_id, fix_value in iteritems(ph_fixed[tree_node._name]):
                        tree_node.fix_variable(variable_id, fix_value)

                # Push fixed variable statuses to instances (or
                # transmit to the phsolverservers)
                ph._push_fix_queue_to_instances()

                # Restore the fix_queue
                for tree_node in ph._scenario_tree._tree_nodes:
                    tree_node._fix_queue.update(
                        ph_fix_queue[tree_node._name])

            elif op == self.TREE_VARIABLE_FIXING:

                ph_fixed, ph_fix_queue = op_data

                # Unfix all non-leaf variables that were fixed to xbar
                for stage in ph._scenario_tree._stages[:-1]:
                    for tree_node in stage._tree_nodes:
                        for variable_id in tree_node._standard_variable_ids:
                            tree_node.free_variable(variable_id)

                # Refix all previously fixed variables
                for tree_node in ph._scenario_tree._tree_nodes:
                    for variable_id, fix_value in iteritems(ph_fixed[tree_node._name]):
                        tree_node.fix_variable(variable_id, fix_value)

                # Push fixed variable statuses to instances (or
                # transmit to the phsolverservers)
                ph._push_fix_queue_to_instances()

                # Restore the fix_queue
                for tree_node in ph._scenario_tree._tree_nodes:
                    tree_node._fix_queue.update(
                        ph_fix_queue[tree_node._name])

            elif op == self.DEACTIVATE_PROXIMAL:

                assert op_data == None
                ph.activate_ph_objective_proximal_terms()

            elif op == self.DEACTIVATE_WEIGHT:

                assert op_data == None
                ph.activate_ph_objective_weight_terms()

            elif op == self.CACHE_WEIGHTS:

                weights = op_data
                for scenario in ph._scenario_tree._scenarios:
                    cached_weights = weights[scenario._name]
                    scenario._w = copy.deepcopy(cached_weights)

            else:

                raise ValueError("Unexpected Operation Flag - "
                                 "This is a developer error")

    def CachePHSolution(self, ph):

        # Cache the current state of the ph instances and scenario
        # tree so we don't affect ph behavior by changing the solution
        # on the scenario tree / scenario instances prior to the next
        # solve
        cache_id = ph.cacheSolutions()

        self._stack.append((self.SOLUTION_CACHING, cache_id))

    def RelaxPHFixedVariables(self, ph):

        # Save the current fixed state and fix queue
        ph_fixed = dict((tree_node._name, copy.deepcopy(tree_node._fixed)) \
                             for tree_node in ph._scenario_tree._tree_nodes)

        ph_fix_queue = \
            dict((tree_node._name, copy.deepcopy(tree_node._fix_queue)) \
                 for tree_node in ph._scenario_tree._tree_nodes)

        # Free all currently fixed variables
        for tree_node in ph._scenario_tree._tree_nodes:
            tree_node.clear_fix_queue()
            for variable_id in ph_fixed[tree_node._name]:
                tree_node.free_variable(variable_id)

        # Push freed variable statuses on instances (or
        # transmit to the phsolverservers)
        ph._push_fix_queue_to_instances()

        self._stack.append((self.VARIABLE_FREEING,
                            (ph_fixed, ph_fix_queue)))

    def FixScenarioTreeVariables(self, ph, fix_values):

        # Save the current fixed state and fix queue and clear the fix queue
        ph_fixed = {}
        ph_fix_queue = {}
        for tree_node in ph._scenario_tree._tree_nodes:
            ph_fixed[tree_node._name] = copy.deepcopy(tree_node._fixed)
            ph_fix_queue[tree_node._name] = copy.deepcopy(tree_node._fix_queue)
            tree_node.clear_fix_queue()

        # Fix everything in fix_values
        for node_name in fix_values:
            tree_node = ph._scenario_tree._tree_node_map[node_name]
            for variable_id, fix_val in iteritems(fix_values[node_name]):
                tree_node.fix_variable(variable_id, fix_val)

        # Push fixed variable statuses on instances (or
        # transmit to the phsolverservers)
        ph._push_fix_queue_to_instances()

        self._stack.append((self.TREE_VARIABLE_FIXING,
                            (ph_fixed, ph_fix_queue)))

    def DeactivatePHObjectiveProximalTerms(self, ph):

        ph.deactivate_ph_objective_proximal_terms()

        self._stack.append((self.DEACTIVATE_PROXIMAL, None))

    def DeactivatePHObjectiveWeightTerms(self, ph):

        ph.deactivate_ph_objective_weight_terms()

        self._stack.append((self.DEACTIVATE_WEIGHT, None))

    def CachePHWeights(self, ph):

        weights = {}
        for scenario in ph._scenario_tree._scenarios:
            weights[scenario._name] = \
                copy.deepcopy(scenario._w)

        self._stack.append((self.CACHE_WEIGHTS,weights))

    #
    # Calculates the probability weighted sum of all suproblem (or
    # bundle) objective functions. This function assumes the current
    # subproblem solutions, in particular the objective values stored
    # on the scenario objects, are appropriate for computing an outer
    # bound on the true optimal objective value (e.g., lower bound for
    # minimization problem). When a nonzero optimality gap is reported
    # in a scenario/bundle solution, the reported objective value for
    # that scenario/bundle will be relaxed accordingly before
    # including it in the average calculation.
    #
    def ComputeOuterBound(self, ph, storage_key):

        if ph._verbose:
            print("Computing objective %s bound" %
                  ("outer" if self._is_minimizing else "inner"))

        bound_status = self.STATUS_NONE
        if (ph._mipgap is not None) and (ph._mipgap > 0):
            logger.warning("A nonzero mipgap was detected when using "
                           "the PH bound plugin. The bound "
                           "computation may as a result be conservative.")
            bound_status |= self.STATUS_MIPGAP

        objective_bound = 0.0

        # If we are bundling, we compute the objective bound in a way
        # such that we can still use solution gap information if it
        # is available.
        if ph._scenario_tree.contains_bundles():

            for scenario_bundle in ph._scenario_tree._scenario_bundles:

                bundle_gap = ph._gaps[scenario_bundle._name]
                bundle_objective_value = 0.0

                for scenario in scenario_bundle._scenario_tree._scenarios:
                    # The objective must be taken from the scenario
                    # objects on PH full scenario tree
                    scenario_objective = \
                        ph._scenario_tree.get_scenario(scenario._name)._objective
                    # And we need to make sure to use the
                    # probabilities assigned to scenarios in the
                    # compressed bundle scenario tree
                    bundle_objective_value += scenario_objective * \
                                              scenario._probability

                if not isinstance(bundle_gap, UndefinedData):
                    if bundle_gap > 0:
                        bound_status |= self.STATUS_GAP_NONZERO
                        if self._is_minimizing:
                            bundle_objective_value -= bundle_gap
                        else:
                            bundle_objective_value += bundle_gap
                else:
                    bound_status |= self.STATUS_GAP_NA

                objective_bound += bundle_objective_value * \
                                   scenario_bundle._probability

        else:

            for scenario in ph._scenario_tree._scenarios:

                this_objective_value = scenario._objective
                this_gap = ph._gaps[scenario._name]

                if not isinstance(this_gap, UndefinedData):
                    if this_gap > 0:
                        bound_status |= self.STATUS_GAP_NONZERO
                        if self._is_minimizing:
                            this_objective_value -= this_gap
                        else:
                            this_objective_value += this_gap
                else:
                    bound_status |= self.STATUS_GAP_NA

                objective_bound += (scenario._probability * this_objective_value)
        if ph._verbose:
            print("Computed objective %s bound=%12.4f\t%s"
                  % (("outer" if self._is_minimizing else "inner"),
                     objective_bound,
                     self.WARNING_MESSAGE.get(bound_status,"")))

        return objective_bound, bound_status

    #
    # Calculates the probability weighted sum of all suproblem (or
    # bundle) objective functions. This function assumes the current
    # subproblem solutions represent valid, non-anticpative solutions
    # are appropriate for computing an inner bound on the optimal
    # objective value (e.g., upper bound for minimization problem).
    #
    def ComputeInnerBound(self, ph, storage_key):

        if ph._verbose:
            print("Computing objective %s bound" %
                  ("inner" if self._is_minimizing else "outer"))

        objective_bound = 0.0
        for scenario in ph._scenario_tree._scenarios:
            objective_bound += (scenario._probability * scenario._objective)

        if ph._verbose:
            print("Computed objective %s bound=%12.4f"
                  % (("inner" if self._is_minimizing else "outer"),
                     objective_bound))

        return objective_bound, self.STATUS_NONE

    def ReportBestBound(self):

        print("")
        best_inner_bound = None
        if len(self._inner_bound_history) > 0: # dlw May 2016
            if self._is_minimizing:
                best_inner_bound = min(self._inner_bound_history.values())
            else:
                best_inner_bound = max(self._inner_bound_history.values())
        print("Best Incumbent Bound: %15s"
              % (best_inner_bound))

        best_bound = None
        if len(self._outer_bound_history) > 0:
            if self._is_minimizing:
                best_bound_key, best_bound = max(self._outer_bound_history.items(),
                                                 key=itemgetter(1))
            else:
                best_bound_key, best_bound = min(self._outer_bound_history.items(),
                                                 key=itemgetter(1))
        print("Best Dual Bound: %15s\t%s"
              % (best_bound,
                 self.WARNING_MESSAGE.get(self._outer_status_history[best_bound_key],"")))

        print("Absolute Duality Gap: %15s"
              % abs(best_inner_bound - best_bound))
        relgap = float('inf')
        if (best_inner_bound != float('inf')) and \
           (best_inner_bound != float('-inf')) and \
           (best_bound != float('-inf')) and \
           (best_bound != float('-inf')):
            relgap = abs(best_inner_bound - best_bound) / \
                     (1e-10+abs(best_bound))

        print("Relative Gap: %15s %s" % (relgap*100.0,"%"))
        print("")
        output_filename = "phbestbound.txt"
        output_file = open(output_filename,"w")
        output_file.write("Incumbent: %.17g\n" % best_inner_bound)
        output_file.write("Dual: %.17g\n" % best_bound)
        output_file.close()
        print("Best bound written to file="+output_filename)

    def ReportBoundHistory(self):
        print("")
        print("Bound History")
        print("%15s %15s %15s" % ("Iteration", "Inner Bound", "Outer Bound"))
        keys = list(self._outer_bound_history.keys())
        if None in keys:
            keys.remove(None)
            print("%15s %15s %15s\t\t%s"
                  % ("Trivial",
                     "       -       ",
                     self._outer_bound_history[None],
                     self.WARNING_MESSAGE.get(self._outer_status_history[None],"")))
        for key in sorted(keys):
            print("%15s %15s %15s\t\t%s"
                  % (key,
                     self._inner_bound_history[key],
                     self._outer_bound_history[key],
                     self.WARNING_MESSAGE.get(self._outer_status_history[key],"")))
        print("")
        output_filename = "phbound.txt"
        with open(output_filename,"w") as output_file:
            output_file.write('Inner Bound:\n')
            for key in sorted(self._inner_bound_history.keys()):
                output_file.write("  %d: %.17g\n"
                                  % (key,
                                     self._inner_bound_history[key]))
            output_file.write('Outer Bound:\n')
            if None in self._outer_bound_history:
                output_file.write("  Trivial: %.17g\n"
                                  % (self._outer_bound_history[None]))
            for key in sorted(keys):
                output_file.write("  %d: %.17g\n"
                                  % (key,
                                     self._outer_bound_history[key]))
        print("Bound history written to file="+output_filename)




#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

__all__ = ('IPySPSolutionSaverExtension',
           'IPySPSolutionLoaderExtension')

from pyomo.common.plugin import Interface

class IPySPSolutionSaverExtension(Interface):

    def save(self, manager):
        """Save solution data from a scenario tree manager.
        Returns True only if data was saved."""
        return False

class IPySPSolutionLoaderExtension(Interface):

    def load(self, manager):
        """Load solution data into a scenario tree manager.
        Returns True only if data was loaded."""
        return True



#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

#
# sorg W: a plugin that cares about the W vectors
#
#
# produce a report of variables sorted by "bad W behavior"
# Major change Sept 2016: only the root node is processed if
#                        flag set
OnlyRootNode = True

import sys

import pyomo.common.plugin
from pyomo.core import *
from pyomo.pysp import phextension
from pyomo.pysp.phutils import *
from pyomo.pysp.generators import \
    scenario_tree_node_variables_generator_noinstances

#==================================================
class sorgwextension(pyomo.common.plugin.SingletonPlugin):

    pyomo.common.plugin.implements(phextension.IPHExtension)

    # the below is a hack to get this extension into the
    # set of IPHExtension objects, so it can be queried
    # automagically by PH.
    pyomo.common.plugin.alias("sorgwextension")

    def __init__(self, *args, **kwds):

        self.Tol = 1e-6
        self.wtrace_filename = 'sorgw.ssv'
        self.wsummary_filename = 'wsummary.ssv'
        self.winterest_filename = 'winterest.ssv' # only vars of interest
        ####### Thresholds for interestingness
        ####### (ored, so any zero causes all to be interesting)
        self.threshWZeroCrossings = 2
        self.threshDiffsRatio = 0.2
        self.threshDiffZeroCrossings = 3

#==================================================
    def reset(self, ph):
        self.__init__()

    def pre_ph_initialization(self,ph):
        # we don't need to intefere with PH initialization.
        pass
#==================================================

#==================================================
    def post_instance_creation(self, ph):
        # we don't muck with the instances.
        pass

#==================================================
    def post_ph_initialization(self, ph):
        pass

#==================================================
    def post_iteration_0_solves(self, ph):
        pass

#==================================================
    def post_iteration_0(self, ph):

        print("sorgw.py is writing the semi-colon separated values file "+self.wtrace_filename)
        ofile = open(self.wtrace_filename, "w")
        self._w_printing(ofile) # write the header
        self._w_printing(ofile, ph)
        ofile.close()

#==================================================

    def pre_iteration_k_solves(self, ph):
        pass

#==================================================
    def post_iteration_k_solves(self, ph):
        pass

#==================================================
    def post_iteration_k(self, ph):

        ofile = open(self.wtrace_filename, "a")
        self._w_printing(ofile, ph)
        ofile.close()

#==================================================
    def post_ph_execution(self, ph):
        # note that we could keep it all in memory and not use a file
        W_Traces = self.Read_W_Traces(self.wtrace_filename)
        self.Compute_and_Write_it_all(W_Traces, ph)

#=========================
    def _w_printing(self, ofile, ph=None):
        # print the w values in a useful way to the open file ofile
        # if ph is None, just write the header
        if ph is None:
            ofile.write("iteration; tree node; scenario; variable; ID; W\n")
        else:
            root_node_name = ph._scenario_tree.findRootNode()._name
            for stage, tree_node, variable_id, variable_values, is_fixed, is_stale in \
                scenario_tree_node_variables_generator_noinstances(ph._scenario_tree,
                                                                   includeDerivedVariables=False,
                                                                   includeLastStage=False):
                if is_stale is False \
                   and (OnlyRootNode is False or tree_node._name == root_node_name):
                    for scenario in tree_node._scenarios:
                       scen_name = scenario._name
                       weight_value = scenario._w[tree_node._name][variable_id]
                       variable_name, index = tree_node._variable_ids[variable_id]
                       full_variable_name = variable_name+indexToString(index)
                       ofile.write(str(ph._current_iteration) + ';' + tree_node._name + ';' + scen_name + ';' + full_variable_name + ';' + str(variable_id)+ ';' + str(weight_value) +'\n')


    #########
    def Read_W_Traces(self, fname):
        # read the W_Traces as written by sorgw.py
        # don't really check for input errors other than a bad iter order

        def num(s):
            try:
                return int(s)
            except ValueError:
                return float(s)

        infile = open(fname,"r")
        curriter = 0
        W_Traces = {}
        for linein in infile:
            parts = linein.split(';')
            if parts[0] == 'iteration': # hack to skip header
                continue
            iternum = int(parts[0])
            nodename = parts[1]
            scenname = parts[2]
            varname = parts[3]
            varid = parts[4]
            wval = num(parts[5])
            if iternum != curriter:
                if iternum-curriter == 1:
                    curriter = iternum
                else:
                    print ("HEY! the input in "+fname+" has iter "+str(iternum)+" after "+str(curriter)+'\n')
                    print (linein)
                    sys.exit(1)
            if varid not in W_Traces:
                W_Traces[varid] = {}
            if scenname not in W_Traces[varid]:
                W_Traces[varid][scenname] = []
            W_Traces[varid][scenname].append(wval)
        return W_Traces

    #####
    def Score_a_Trace(self, wtrace):
        # given a list of w values, compute and return scores
        # (lower is better)
        def dlwmean(v):
            if len(v) > 0:
                return sum(v)/len(v)
            else:
                return 0
        WZeroCrossings = 0  # a score
        WAbove = False
        WBelow = False
        # note: hitting zero does not reset Above or Below or cause a crossing
        for w in wtrace:
            if abs(w) < self.Tol:
                continue
            if w > self.Tol:
                if WBelow:
                    WZeroCrossings += 1
                    WBelow = False
                WAbove = True
            if w < self.Tol:
                if WAbove:
                    WZeroCrossings += 1
                    WAbove = False
                WBelow = True

        DiffsRatio = 0  # a score
        DiffZeroCrossings = 0  # a score
        absdiffs = [] # t to t+1
        DiffAbove = False
        DiffBelow = False
        wcnt = len(wtrace)
        for i in range(wcnt-1):
            diff = wtrace[i+1] - wtrace[i]
            absdiffs.append(abs(diff))
            if abs(diff) < self.Tol:
                continue
            if diff > self.Tol:
                if DiffBelow:
                    DiffZeroCrossings += 1
                    DiffBelow = False
                DiffAbove = True
            if diff < self.Tol:
                if DiffAbove:
                    DiffZeroCrossings += 1
                    DiffAbove = False
                DiffBelow = True
        frontavg = dlwmean(absdiffs[:int(wcnt/2)])
        if frontavg > self.Tol:
            DiffsRatio = dlwmean(absdiffs[int(wcnt/2):]) / frontavg # close enough
        return WZeroCrossings, DiffsRatio, DiffZeroCrossings

    ####
    def Of_Interest(self, WZeroCrossings, DiffsRatio, DiffZeroCrossings):
    # return true if anything is above its threshold
        return WZeroCrossings >= self.threshWZeroCrossings or \
               DiffsRatio >= self.threshDiffsRatio or \
               DiffZeroCrossings >= self.threshDiffZeroCrossings

    ####
    def Compute_and_Write_it_all(self, W_Traces, ph):
        VarsOfInterest = set()
        fname = self.wsummary_filename
        print ("sorgw.py is writing the semi-colon separated values file "+fname)
        ofile = open(fname, "w")
        ofile.write("var; scen; WZeroCrossing; DiffsRatio; DiffZeroCrossings; w values...\n")
        for varid in W_Traces:
            assert(OnlyRootNode)
            variable_name, index = ph._scenario_tree.findRootNode()._variable_ids[varid]
            varname = variable_name+indexToString(index)

            for scenname in W_Traces[varid]:
                WZeroCrossings, DiffsRatio, DiffZeroCrossings = self.Score_a_Trace(W_Traces[varid][scenname])
                if self.Of_Interest(WZeroCrossings, DiffsRatio, DiffZeroCrossings):
                    VarsOfInterest.add(varid)
                ofile.write(varname+';'+scenname+';'+str(WZeroCrossings)+';'+str(DiffsRatio)+';'+str(DiffZeroCrossings))
                for w in W_Traces[varid][scenname]:
                    ofile.write(';'+str(w))
                ofile.write('\n')
        ofile.close
        # now processing interesting vars
        BiggestLoser = None
        LoserRange = 0
        fname = self.winterest_filename
        print ("sorgw.py is writing the semi-colon separated values file "+fname)
        ofile = open(fname, "w")
        ofile.write("var; scen; WZeroCrossing; DiffsRatio; DiffZeroCrossings; w values...\n")
        for varid in VarsOfInterest:
            vwmax = vwmin = 0
            assert(OnlyRootNode)
            variable_name, index = ph._scenario_tree.findRootNode()._variable_ids[varid]
            varname = variable_name+indexToString(index)

            for scenname in W_Traces[varid]:
                WZeroCrossings, DiffsRatio, DiffZeroCrossings = self.Score_a_Trace(W_Traces[varid][scenname])
                ofile.write(varname+';'+scenname+';'+str(WZeroCrossings)+';'+str(DiffsRatio)+';'+str(DiffZeroCrossings))
                for w in W_Traces[varid][scenname]:
                    ofile.write(';'+str(w))
                    if w > vwmax:
                        vwmax = w
                    if w < vwmin:
                        vwmin = w
                ofile.write('\n')
            vwrange = vwmax - vwmin
            if vwrange > LoserRange:
                LoserRange = vwrange
                BiggestLoser = varid

        ofile.close
        if BiggestLoser is not None:
            ph._sorgw_BiggestLoser = BiggestLoser
        print ("sorgw.py complete: RootNodeOnly="+str(OnlyRootNode)+" BiggestLoser="+str(BiggestLoser)+" with vwrange="+str(vwrange))



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

from pyomo.core.base import *
from pyomo.opt import (ProblemFormat,
                       SolverFactory,
                       SolverManagerFactory,
                       SolverStatus,
                       TerminationCondition,
                       SolutionStatus)
from pyomo.pysp.phutils import (isVariableNameIndexed,
                                extractVariableNameAndIndex,
                                extractComponentIndices)

#
# a routine to create the extensive form, given an input scenario tree and instances.
# IMPT: unlike scenario instances, the extensive form instance is *not* self-contained.
#       in particular, it has binding constraints that cross the binding instance and
#       the scenario instances. it is up to the caller to keep track of which scenario
#       instances are associated with the extensive form. this might be something we
#       encapsulate at some later time.

#
# GAH: The following comment is still true, except that the scenario tree is no longer modified
#
# NOTE: if cvar terms are generated, then the input scenario tree is modified accordingly,
#       i.e., with the addition of the "eta" variable at the root node and the excess
#       variables at (for lack of a better place - they are per-scenario, but are not
#       blended) the second stage.
#

def create_ef_instance(scenario_tree,
                       ef_instance_name = "MASTER",
                       verbose_output = False,
                       generate_weighted_cvar = False,
                       cvar_weight = None,
                       risk_alpha = None,
                       cc_indicator_var_name=None,
                       cc_alpha=0.0):

    #
    # create the new and empty binding instance.
    #

    # scenario tree must be "linked" with a set of instances
    # to used this function
    scenario_instances = {}
    for scenario in scenario_tree.scenarios:
        if scenario._instance is None:
            raise ValueError(
                "Cannot construct extensive form instance. "
                "The scenario tree does not appear to be linked "
                "to any Pyomo models. Missing model for scenario "
                "with name: %s" % (scenario.name))
        scenario_instances[scenario.name] = scenario._instance

    binding_instance = ConcreteModel(name=ef_instance_name)
    root_node = scenario_tree.findRootNode()

    opt_sense = minimize \
                if (scenario_tree._scenarios[0]._instance_objective.is_minimizing()) \
                   else maximize

    #
    # validate cvar options, if specified.
    #
    cvar_excess_vardatas = []
    if generate_weighted_cvar:
        if (cvar_weight is None) or (cvar_weight < 0.0):
            raise RuntimeError("Weight of CVaR term must be >= 0.0 - value supplied="+str(cvar_weight))
        if (risk_alpha is None) or (risk_alpha <= 0.0) or (risk_alpha >= 1.0):
            raise RuntimeError("CVaR risk alpha must be between 0 and 1, exclusive - value supplied="+str(risk_alpha))

        if verbose_output:
            print("Writing CVaR weighted objective")
            print("CVaR term weight="+str(cvar_weight))
            print("CVaR alpha="+str(risk_alpha))
            print("")

        # create the eta and excess variable on a per-scenario basis,
        # in addition to the constraint relating to the two.

        cvar_eta_variable_name = "CVAR_ETA_"+str(root_node._name)
        cvar_eta_variable = Var()
        binding_instance.add_component(cvar_eta_variable_name, cvar_eta_variable)

        excess_var_domain = NonNegativeReals if (opt_sense == minimize) else \
                            NonPositiveReals

        compute_excess_constraint = \
            binding_instance.COMPUTE_SCENARIO_EXCESS = \
                ConstraintList()

        for scenario in scenario_tree._scenarios:

            cvar_excess_variable_name = "CVAR_EXCESS_"+scenario._name
            cvar_excess_variable = Var(domain=excess_var_domain)
            binding_instance.add_component(cvar_excess_variable_name, cvar_excess_variable)

            compute_excess_expression = cvar_excess_variable
            compute_excess_expression -= scenario._instance_cost_expression
            compute_excess_expression += cvar_eta_variable
            if opt_sense == maximize:
                compute_excess_expression *= -1

            compute_excess_constraint.add((0.0, compute_excess_expression, None))

            cvar_excess_vardatas.append((cvar_excess_variable, scenario._probability))

    # the individual scenario instances are sub-blocks of the binding instance.
    for scenario in scenario_tree._scenarios:
        scenario_instance = scenario_instances[scenario._name]
        binding_instance.add_component(str(scenario._name), scenario_instance)
        # Now deactivate the scenario instance Objective since we are creating
        # a new master objective
        scenario._instance_objective.deactivate()

    # walk the scenario tree - create variables representing the
    # common values for all scenarios associated with that node, along
    # with equality constraints to enforce non-anticipativity.  also
    # create expected cost variables for each node, to be computed via
    # constraints/objectives defined in a subsequent pass. master
    # variables are created for all nodes but those in the last
    # stage. expected cost variables are, for no particularly good
    # reason other than easy coding, created for nodes in all stages.
    if verbose_output:
        print("Creating variables for master binding instance")

    _cmap = binding_instance.MASTER_CONSTRAINT_MAP = ComponentMap()
    for stage in scenario_tree._stages[:-1]: # skip the leaf stage

        for tree_node in stage._tree_nodes:

            # create the master blending variable and constraints for this node
            master_blend_variable_name = \
                "MASTER_BLEND_VAR_"+str(tree_node._name)
            master_blend_constraint_name = \
                "MASTER_BLEND_CONSTRAINT_"+str(tree_node._name)

            # don't create master variables for derived
            # stage variables as they will not be used in
            # the problem, and their values would likely
            # never be consistent with what is stored on the
            # scenario variables
            master_variable_index = Set(initialize=sorted(tree_node._standard_variable_ids),
                                        ordered=True,
                                        name=master_blend_variable_name+"_index")

            binding_instance.add_component(master_blend_variable_name+"_index",
                                           master_variable_index)

            master_variable = Var(master_variable_index,
                                  name=master_blend_variable_name)

            binding_instance.add_component(master_blend_variable_name,
                                           master_variable)

            master_constraint = ConstraintList(name=master_blend_constraint_name)

            binding_instance.add_component(master_blend_constraint_name,
                                           master_constraint)

            tree_node_variable_datas = tree_node._variable_datas
            for variable_id in sorted(tree_node._standard_variable_ids):
                master_vardata = master_variable[variable_id]
                vardatas = tree_node_variable_datas[variable_id]
                # Don't blend fixed variables
                if not tree_node.is_variable_fixed(variable_id):
                    for scenario_vardata, scenario_probability in vardatas:
                        _cmap[scenario_vardata] = master_constraint.add(
                            (master_vardata-scenario_vardata,0.0))

    if generate_weighted_cvar:

        cvar_cost_expression_name = "CVAR_COST_"+str(root_node._name)
        cvar_cost_expression = Expression(name=cvar_cost_expression_name)
        binding_instance.add_component(cvar_cost_expression_name, cvar_cost_expression)

    # create an expression to represent the expected cost at the root node
    binding_instance.EF_EXPECTED_COST = \
        Expression(initialize=sum(scenario._probability * \
                                  scenario._instance_cost_expression
                                  for scenario in scenario_tree._scenarios))

    opt_expression = \
        binding_instance.MASTER_OBJECTIVE_EXPRESSION = \
            Expression(initialize=binding_instance.EF_EXPECTED_COST)

    if generate_weighted_cvar:
        cvar_cost_expression_name = "CVAR_COST_"+str(root_node._name)
        cvar_cost_expression = \
            binding_instance.find_component(cvar_cost_expression_name)
        if cvar_weight == 0.0:
            # if the cvar weight is 0, then we're only
            # doing cvar - no mean.
            opt_expression.set_value(cvar_cost_expression)
        else:
            opt_expression.expr += cvar_weight * cvar_cost_expression

    binding_instance.MASTER = Objective(sense=opt_sense,
                                        expr=opt_expression)

    # CVaR requires the addition of a variable per scenario to
    # represent the cost excess, and a constraint to compute the cost
    # excess relative to eta.
    if generate_weighted_cvar:

        # add the constraint to compute the master CVaR variable value. iterate
        # over scenario instances to create the expected excess component first.
        cvar_cost_expression_name = "CVAR_COST_"+str(root_node._name)
        cvar_cost_expression = binding_instance.find_component(cvar_cost_expression_name)
        cvar_eta_variable_name = "CVAR_ETA_"+str(root_node._name)
        cvar_eta_variable = binding_instance.find_component(cvar_eta_variable_name)

        cost_expr = 1.0
        for scenario_excess_vardata, scenario_probability in cvar_excess_vardatas:
            cost_expr += (scenario_probability * scenario_excess_vardata)
        cost_expr /= (1.0 - risk_alpha)
        cost_expr += cvar_eta_variable

        cvar_cost_expression.set_value(cost_expr)

    if cc_indicator_var_name is not None:
        if verbose_output is True:
            print("Creating chance constraint for indicator variable= "+cc_indicator_var_name)
            print( "with alpha= "+str(cc_alpha))
        if not isVariableNameIndexed(cc_indicator_var_name):
            cc_expression = 0  #??????
            for scenario in scenario_tree._scenarios:
                scenario_instance = scenario_instances[scenario._name]
                scenario_probability = scenario._probability
                cc_var = scenario_instance.find_component(cc_indicator_var_name)

                cc_expression += scenario_probability * cc_var

            def makeCCRule(expression):
                def CCrule(model):
                    return(1.0 - cc_alpha, cc_expression, None)
                return CCrule

            cc_constraint_name = "cc_"+cc_indicator_var_name
            cc_constraint = Constraint(name=cc_constraint_name, rule=makeCCRule(cc_expression))
            binding_instance.add_component(cc_constraint_name, cc_constraint)
        else:
            print("multiple cc not yet supported.")
            variable_name, index_template = extractVariableNameAndIndex(cc_indicator_var_name)

            # verify that the root variable exists and grab it.
            # NOTE: we are using whatever scenario happens to laying around... it might be better to use the reference
            variable = scenario_instance.find_component(variable_name)
            if variable is None:
                raise RuntimeError("Unknown variable="+variable_name+" referenced as the CC indicator variable.")

            # extract all "real", i.e., fully specified, indices matching the index template.
            match_indices = extractComponentIndices(variable, index_template)

            # there is a possibility that no indices match the input template.
            # if so, let the user know about it.
            if len(match_indices) == 0:
                raise RuntimeError("No indices match template="+str(index_template)+" for variable="+variable_name)

            # add the suffix to all variable values identified.
            for index in match_indices:
                variable_value = variable[index]

                cc_expression = 0  #??????
                for scenario in scenario_tree._scenarios:
                    scenario_instance = scenario_instances[scenario._name]
                    scenario_probability = scenario._probability
                    cc_var = scenario_instance.find_component(variable_name)[index]

                    cc_expression += scenario_probability * cc_var

                def makeCCRule(expression):
                    def CCrule(model):
                        return(1.0 - cc_alpha, cc_expression, None)
                    return CCrule

                indexasname = ''
                for c in str(index):
                   if c not in ' ,':
                      indexasname += c
                cc_constraint_name = "cc_"+variable_name+"_"+indexasname

                cc_constraint = Constraint(name=cc_constraint_name, rule=makeCCRule(cc_expression))
                binding_instance.add_component(cc_constraint_name, cc_constraint)

    return binding_instance

#
# write the EF binding instance and all sub-instances.
#

def write_ef(binding_instance,
             output_filename,
             symbolic_solver_labels=False,
             output_fixed_variable_bounds=False):

    # determine the output file type, and invoke the appropriate
    # writer.
    pieces = output_filename.rsplit(".",1)
    if len(pieces) != 2:
        raise RuntimeError("Could not determine suffix from "
                           "output filename="+output_filename)
    ef_output_file_suffix = pieces[1]

    io_options = {}
    if symbolic_solver_labels:
        io_options['symbolic_solver_labels'] = True
    if output_fixed_variable_bounds:
        io_options['output_fixed_variable_bounds'] = True

    # create the output file.
    if ef_output_file_suffix == "lp":

        _, smap_id = binding_instance.write(
            filename=output_filename,
            format=ProblemFormat.cpxlp,
            solver_capability=lambda x: True,
            io_options=io_options)

    elif ef_output_file_suffix == "nl":

        _, smap_id = binding_instance.write(
            filename=output_filename,
            format=ProblemFormat.nl,
            solver_capability=lambda x: True,
            io_options=io_options)

    elif ef_output_file_suffix == "mps":

        _, smap_id = binding_instance.write(
            filename=output_filename,
            format=ProblemFormat.mps,
            solver_capability=lambda x: True,
            io_options=io_options)

    else:
        raise RuntimeError("Unknown file suffix="+ef_output_file_suffix+
                           " specified when writing extensive form")

    return smap_id

#
# solve the EF binding instance and load the solution
#

def solve_ef(master_instance, options):

    with SolverFactory(options.solver_type) as ef_solver:

        with SolverManagerFactory(options.solver_manager_type) as ef_solver_manager:
            if ef_solver is None:
                raise ValueError("Failed to create solver manager of type="
                                 +options.solver_type+" for use in extensive form solve")
            if len(options.ef_solver_options) > 0:
                print("Initializing ef solver with options="+str(options.ef_solver_options))
                ef_solver.set_options("".join(options.ef_solver_options))
            if options.ef_mipgap is not None:
                if (options.ef_mipgap < 0.0) or \
                   (options.ef_mipgap > 1.0):
                    raise ValueError("Value of the mipgap parameter for the "
                                     "EF solve must be on the unit interval; "
                                     "value specified="+str(options.ef_mipgap))
                else:
                    ef_solver.mipgap = options.ef_mipgap

            solve_kwds = {}
            solve_kwds['load_solutions'] = False
            if options.keep_solver_files:
                solve_kwds['keepfiles'] = True
            if options.symbolic_solver_labels:
                solve_kwds['symbolic_solver_labels'] = True
            if options.output_solver_log:
                solve_kwds['tee'] = True
            if options.write_fixed_variables:
                solve_kwds['output_fixed_variable_bounds'] = True

            if options.verbose:
                print("Solving extensive form.")

            if (not options.disable_warmstarts) and \
               (ef_solver.warm_start_capable()):
                ef_action_handle = ef_solver_manager.queue(
                    master_instance,
                    opt=ef_solver,
                    warmstart=True,
                    **solve_kwds)
            else:
                ef_action_handle = ef_solver_manager.queue(
                    master_instance,
                    opt=ef_solver,
                    **solve_kwds)
            results = ef_solver_manager.wait_for(ef_action_handle)

            # check the return code - if this is anything but "have a solution", we need to bail.
            if (results.solver.status == SolverStatus.ok) and \
               ((results.solver.termination_condition == TerminationCondition.optimal) or \
                ((len(results.solution) > 0) and (results.solution(0).status == SolutionStatus.optimal))):

                master_instance.solutions.load_from(results)
                return results

    raise RuntimeError("Extensive form was infeasible!")



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
#

import six
import pickle
import base64
import ast
import os
import sys
from os.path import abspath, dirname

from pyutilib.pyro import using_pyro4
import pyutilib.th as unittest
import pyutilib.services
from pyutilib.misc import Options
import pyomo.opt
from pyomo.environ import *
import pyomo.scripting.pyro_mip_server

import six

currdir = dirname(abspath(__file__))+os.sep

solvers = pyomo.opt.check_available_solvers('glpk')

class TestWorker(pyomo.scripting.pyro_mip_server.PyomoMIPWorker):

    def __init__(self):
        self._verbose = True
        self._current_task_client = None

class Test(unittest.TestCase):

    def setUp(self):
        self.worker = TestWorker()

    def tearDown(self):
        pyutilib.services.TempfileManager.clear_tempfiles()
        del self.worker

    @unittest.skipIf(not 'glpk' in solvers, "glpk solver is not available")
    def test_t1(self):
        # Run a simple model
        model = ConcreteModel()
        model.A = RangeSet(1,4)
        model.x = Var(model.A, bounds=(-1,1))
        def obj_rule(model):
            return sum_product(model.x)
        model.obj = Objective(rule=obj_rule)
        def c_rule(model):
            expr = 0
            for i in model.A:
                expr += i*model.x[i]
            return expr == 0
        model.c = Constraint(rule=c_rule)

        #
        data = Options()
        data.suffixes = {}
        data.solver_options = {}
        data.warmstart_filename = None
        data.filename = currdir+'t1.lp'
        model.write(data['filename'])
        INPUT = open(data['filename'],'r')
        data['file'] = INPUT.read()
        INPUT.close()
        data['opt'] = 'glpk'
        data.kwds = {}
        #
        results = self.worker.process(data)

        # Decode, evaluate and unpickle results
        if using_pyro4:
            # These two conversions are in place to unwrap
            # the hacks placed in the pyro_mip_server
            # before transmitting the results
            # object. These hacks are put in place to
            # avoid errors when transmitting the pickled
            # form of the results object with the default Pyro4
            # serializer (Serpent)
            if six.PY3:
                results = base64.decodebytes(
                    ast.literal_eval(results))
            else:
                results = base64.decodestring(results)

        results = pickle.loads(results)

        #
        results.write(filename=currdir+"t1.out", format='json')
        self.assertMatchesJsonBaseline(currdir+"t1.out",currdir+"t1.txt", tolerance=1e-4)
        self.assertEqual(results._smap_id, None)
        os.remove(data['filename'])


if __name__ == "__main__":
    unittest.main()




#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

# TODO: have smps convert create symbols files for second stage
# TODO: parse second-stage solution and load into scenario tree workers
# TODO: objective, cost, stage_costs
# TODO: make output_scenario_tree_solution work
# TODO: results object statuses

# TODO Fix the seed defaults. The I think the default
#      behavior should be to provide no needs and enable
#      AUTO_SEED. There are so many seeds to provide that
#      I'm not sure where to start on this. Perhaps, the
#      SD developers can updates these defaults. It would be
#      nice if a single seed could be provided and the rest
#      generated from this by SD, when deterministic behavior
#      is required.

import os
import sys
import time
import traceback
import logging

import pyutilib.subprocess
import pyutilib.services

from pyomo.core import maximize
from pyomo.opt import (TerminationCondition,
                       SolverStatus,
                       SolutionStatus)
from pyomo.pysp.util.configured_object import PySPConfiguredObject
from pyomo.pysp.util.config import (PySPConfigValue,
                                    PySPConfigBlock,
                                    safe_register_common_option,
                                    safe_register_unique_option,
                                    safe_declare_common_option,
                                    safe_declare_unique_option,
                                    _domain_positive,
                                    _domain_positive_integer,
                                    _domain_nonnegative_integer,
                                    _domain_must_be_str)
from pyomo.pysp.util.misc import (parse_command_line,
                                  launch_command)
from pyomo.pysp.scenariotree.manager import \
    ScenarioTreeManagerFactory
import pyomo.pysp.convert.smps
from pyomo.pysp.embeddedsp import EmbeddedSP
from pyomo.pysp.solvers.spsolver import (SPSolverResults,
                                         SPSolverFactory)
from pyomo.pysp.solvers.spsolvershellcommand import \
    SPSolverShellCommand

logger = logging.getLogger('pyomo.pysp')

_sd_group_label = "SD Options"

# maps the sd status message to tuples of the form
# (SolutionStatus, SolverStatus, TerminationCondition)
# feel free to modify if you have opinions one these
_sd_status_map = {}
_sd_status_map["COMPROMISE SOLUTION"] = (SolutionStatus.optimal,
                                         SolverStatus.ok,
                                         TerminationCondition.optimal)
_sd_status_map["MEAN SOLUTION"] = (SolutionStatus.optimal,
                                   SolverStatus.ok,
                                   TerminationCondition.optimal)
_sd_status_map["REPLICATION SOLUTION"] = (SolutionStatus.optimal,
                                          SolverStatus.ok,
                                          TerminationCondition.optimal)
_sd_status_map["NO  SOLUTION"] = (SolutionStatus.unknown,
                                  SolverStatus.warning,
                                  TerminationCondition.noSolution)
_sd_status_map["???"] = (SolutionStatus.unknown,
                         SolverStatus.error,
                         TerminationCondition.unknown)

sd_advanced_config_section = \
"""
// ---------------Changes made below here are not recommended-----------------


// 0 for LP master, 1 for QP master
MASTER_TYPE 1

// 0 for zero lower bound, 1 for mean value lower bound
LB_TYPE 1

// Amount of improvement which must be observed in order to update incumbent X.
R 0.2

// For updating the scaling factor. zl
R2 0.95

// For updating the scaling factor. zl
R3 2.0

// The Minimum value of the cell->quad_scalar. zl
MIN_QUAD_SCALAR 0.001

// The Maximum value of the cell->quad_scalar. zl
MAX_QUAD_SCALAR 10000.0

// Number of iterations after which incumbent cut is reevaluated.
TAU 2

// Ratio used in pre_test for optimality: (Sm - Lm) / Fk  < PRE_EPSILON.
PRE_EPSILON 0.01

// MIN_ITER will be max(ITER_FACT*xdim, MIN_ITER)   JH 3/13/98
ITER_FACT 0

// Percent of resampled solutions which must meet test for optimality.
PERCENT_PASS 0.95

// Number of resampling to take when checking optimality.
M 50

// Multiplier for the number of cuts; with QP, SD sets this to 1
CUT_MULT  5

// Level of confidence when choosing the set of cuts, T, for optimality test.
CONFID_HI 1.0

// Level of confidence when conducting the optimality pre-test.
CONFID_LO 1.45

// Percent of the number by which two "equal" numbers may differ.
TOLERANCE 0.001

// number by which two "equal" numbers may differ (used in judging feasibility).
FEA_TOLER 0.05

// Like tolerance, but used when trying to thin omegas.
THIN_TOLER 0.001

// Number of iterations before SD tries to thin the data structures.
START_THIN 9001

// Number of iterations between successive thinning of structures.
THIN_CYCLE 200


// Number of consecutive iterations a cut must be slack before its dropped.
DROP_TIME 16300

// 0 for pre-tests only, 1 for full_test
TEST_TYPE 1

// Number of iterations before checking new pi's impact
PI_EVAL_START 1

// Print out detailed solutions, set this to 1
DETAILED_SOLN 1

// Number of iterations between evaluation of the impact of new dual vertex
PI_CYCLE 1

// The flag for subproblem lower bound checker.
// 0 -- no subproblem LB checking.
// 1 -- check subproblem LB.
SUB_LB_CHECK 0

// The flag for seed generation method
// 0 -- user provide seeds manually
// 1 -- SD generates seeds automatically
AUTO_SEED  0

// 16 digits are recommended for the seed
// Random number seed for generating observations of omega.
// RUN_SEED1     9495518635394380
RUN_SEED1     9495518635394380
RUN_SEED2     4650175399072632
RUN_SEED3     6070772756632709
RUN_SEED4     5451675876709589
RUN_SEED5     5285327724846206
RUN_SEED6     5588857889468088
RUN_SEED7     1098833779416153
RUN_SEED8     6192593982049265
RUN_SEED9     4756774140130874
RUN_SEED10    6784592265109609
RUN_SEED11    9728429908537680
RUN_SEED12    1163479388309571
RUN_SEED13    3279282318700126
RUN_SEED14    8773753208032360
RUN_SEED15    9337302665697748
RUN_SEED16    4415169667296773
RUN_SEED17    4220432037464045
RUN_SEED18    3554548844580680
RUN_SEED19    1814300451929103
RUN_SEED20    5339672949292608
RUN_SEED21    5638710736762732
RUN_SEED22    3154245808720589
RUN_SEED23    2414929536171258
RUN_SEED24    7998609999427572
RUN_SEED25    7080145164625719
RUN_SEED26    3648862740490586
RUN_SEED27    7772725003305823
RUN_SEED28    5982768791029230
RUN_SEED29    1395182510837913
RUN_SEED30    3735836402047426
"""

def _domain_sd_tolerance(val):
    val = str(val)
    if val not in _domain_sd_tolerance._values:
        raise ValueError(
            "Value %s is not one of %s."
            % (val, str(_domain_sd_tolerance._values)))
    return val
_domain_sd_tolerance._values = ('loose','nominal','tight')
_domain_sd_tolerance.doc = \
    "<domain: %s>" % (str(_domain_sd_tolerance._values))

class SDSolver(SPSolverShellCommand, PySPConfiguredObject):

    @classmethod
    def _declare_options(cls, options=None):
        if options is None:
            options = PySPConfigBlock()
        safe_declare_unique_option(
            options,
            "stopping_rule_tolerance",
            PySPConfigValue(
                "nominal",
                domain=_domain_sd_tolerance,
                description=(
                    "Stopping rule tolerance used by the SD solver. "
                    "Must be one of: %s. Default is 'nominal'."
                    % (str(_domain_sd_tolerance._values))
                ),
                doc=None,
                visibility=0),
            ap_group=_sd_group_label)
        safe_declare_unique_option(
            options,
            "single_replication",
            PySPConfigValue(
                False,
                domain=bool,
                description=(
                    "Disables multiple replication procedure in "
                    "SD and uses a single replication."
                ),
                doc=None,
                visibility=0),
            ap_group=_sd_group_label)
        safe_declare_unique_option(
            options,
            "print_cycle",
            PySPConfigValue(
                100,
                domain=_domain_positive_integer,
                description=(
                    "Number of iterations between output of "
                    "solution data to screen and file."
                ),
                doc=None,
                visibility=0),
            ap_group=_sd_group_label)
        safe_declare_unique_option(
            options,
            "eval_run_flag",
            PySPConfigValue(
                False,
                domain=bool,
                description=(
                    "Set to evaluate on the run. This should be "
                    "only used for instances with relatively complete "
                    "recourse. This flag is not recommended because "
                    "accurate function evaluations are unnecessarily "
                    "time consuming. It is best to use a large print "
                    "cycle when this option is activated."
                ),
                doc=None,
                visibility=0),
            ap_group=_sd_group_label)
        safe_declare_unique_option(
            options,
            "eval_flag",
            PySPConfigValue(
                False,
                domain=bool,
                description=(
                    "Set to get an estimated objective function value "
                    "for the final incumbent of each replication."
                ),
                doc=None,
                visibility=0),
            ap_group=_sd_group_label)
        safe_declare_unique_option(
            options,
            "eval_seed1",
            PySPConfigValue(
                2668655841019641,
                domain=int,
                description=(
                    "Random number seed for re-sampling omegas during "
                    "optimality test. Default is None, meaning no "
                    "seed will be provided."
                ),
                doc=None,
                visibility=0),
            ap_group=_sd_group_label)
        safe_declare_unique_option(
            options,
            "eval_error",
            PySPConfigValue(
                0.01,
                domain=_domain_positive,
                description=(
                    "Objective evaluation is accurate to within "
                    "this much, with 95%% confidence. Default is 0.01."
                ),
                doc=None,
                visibility=0),
            ap_group=_sd_group_label)
        safe_declare_unique_option(
            options,
            "mean_dev",
            PySPConfigValue(
                0.05,
                domain=_domain_positive,
                description=(
                    "Solution tolerance for deciding the usage of "
                    "mean solution. Default is 0.05."
                ),
                doc=None,
                visibility=0),
            ap_group=_sd_group_label)
        safe_declare_unique_option(
            options,
            "min_iterations",
            PySPConfigValue(
                None,
                domain=_domain_nonnegative_integer,
                description=(
                    "Number of iterations which must pass before "
                    "optimality is checked. Default is None, meaning "
                    "no minimum is given."
                ),
                doc=None,
                visibility=0),
            ap_group=_sd_group_label)
        safe_declare_unique_option(
            options,
            "max_iterations",
            PySPConfigValue(
                5000,
                domain=_domain_positive_integer,
                description=(
                    "Maximum number of iterations for any given "
                    "problem. Default is 5000."
                ),
                doc=None,
                visibility=0),
            ap_group=_sd_group_label)
        safe_declare_common_option(options,
                                   "verbose",
                                   ap_group=_sd_group_label)

        return options

    def __init__(self):
        super(SDSolver, self).__init__(self.register_options())
        self.set_options_to_default()
        self._executable = "sd"

    def set_options_to_default(self):
        self._options = self.register_options()

    @property
    def options(self):
        return self._options

    @property
    def name(self):
        return "sd"

    def _solve_impl(self,
                    sp,
                    output_solver_log=False,
                    **kwds):
        """
        Solve a stochastic program with the SD solver.

        See the 'solve' method on the base class for
        additional keyword documentation.

        Args:
            sp: The stochastic program to solve.
            output_solver_log (bool): Stream the solver
                output during the solve.
            **kwds: Passed to the SMPS file writer as I/O
              options (e.g., symbolic_solver_labels=True).

        Returns: A results object with information about the solution.
        """

        if sp.objective_sense == maximize:
            raise ValueError("SD solver does not yet handle "
                             "maximization problems")

        #
        # Setup the SD working directory
        #

        working_directory = self._create_tempdir("workdir")
        config_filename = self._files["config"] = \
                          os.path.join(working_directory,
                                       "config.sd")
        self._add_tempfile("config", config_filename)
        sdinput_directory = os.path.join(working_directory,
                                         "sdinput",
                                         "pysp_model")
        self._add_tempfile("stdinput", sdinput_directory)
        sdoutput_directory = os.path.join(working_directory,
                                          "sdoutput",
                                          "pysp_model")
        self._add_tempfile("sdoutput", sdoutput_directory)
        logfile = os.path.join(working_directory, "sd.log")
        self._add_tempfile("log", logfile)

        if self.get_option("verbose"):
            print("Writing solver files in directory: %s"
                  % (working_directory))

        os.makedirs(sdinput_directory)
        assert os.path.exists(sdinput_directory)
        assert not os.path.exists(sdoutput_directory)
        self._write_config(config_filename)

        if self.get_option('single_replication'):
            solution_filename = os.path.join(
                sdoutput_directory,
                "pysp_model.detailed_rep_soln.out")
        else:
            solution_filename = os.path.join(
                sdoutput_directory,
                "pysp_model.detailed_soln.out")
        self._add_tempfile("solution", solution_filename)

        #
        # Create the SD input files
        #

        if isinstance(sp, EmbeddedSP):
            input_files, _ = pyomo.pysp.convert.smps.\
                convert_embedded(
                    sdinput_directory,
                    "pysp_model",
                    sp,
                    core_format='mps',
                    io_options=kwds)
        else:
            input_files = pyomo.pysp.convert.smps.\
                convert_external(
                    sdinput_directory,
                    "pysp_model",
                    sp,
                    core_format='mps',
                    io_options=kwds)
        for key in input_files:
            self._add_tempfile(key, input_files[key])

        #
        # Launch SD
        #

        _cmd_string = self.executable+" < pysp_model"
        if self.get_option("verbose"):
            print("Launching SD solver with command: %s"
                  % (_cmd_string))

        start = time.time()
        rc, log = pyutilib.subprocess.run(
            self.executable,
            cwd=working_directory,
            stdin="pysp_model",
            outfile=logfile,
            tee=output_solver_log)
        stop = time.time()
        assert os.path.exists(sdoutput_directory)

        #
        # Parse the SD solution
        #

        if self.get_option("verbose"):
            print("Reading SD solution from file: %s"
                  % (solution_filename))

        xhat, results = self._read_solution(input_files["symbols"],
                                            solution_filename)
        results.solver.time = stop - start

        results.xhat = None
        if xhat is not None:
            if isinstance(sp, EmbeddedSP):
                results.xhat = {sp.time_stages[0]: xhat}
            else:
                results.xhat = {sp.scenario_tree.findRootNode().name: xhat}

        return results

    def _write_config(self, filename):
        """ Writes an SD config file """
        with open(filename, "w") as f:
            if self.get_option("stopping_rule_tolerance") == "loose":
                f.write("// nominal tolerance\n")
                f.write("EPSILON 0.01\n")
                f.write("SCAN_LEN 64\n")
            elif self.get_option("stopping_rule_tolerance") == "nominal":
                f.write("// nominal tolerance\n")
                f.write("EPSILON 0.001\n")
                f.write("SCAN_LEN 256\n")
            elif self.get_option("stopping_rule_tolerance") == "tight":
                f.write("// nominal tolerance\n")
                f.write("EPSILON 0.00001\n")
                f.write("SCAN_LEN 512\n")
            else:
                assert False
            f.write("MULTIPLE_REP %d\n"
                    % (0 if self.get_option("single_replication") else 1))
            f.write("PRINT_CYCLE %d\n" % (self.get_option("print_cycle")))
            f.write("EVAL_RUN_FLAG %d\n"
                    % (1 if self.get_option("eval_run_flag") else 0))
            f.write("EVAL_FLAG %d\n"
                    % (1 if self.get_option("eval_flag") else 0))
            f.write("EVAL_SEED1 %d\n" % (self.get_option("eval_seed1")))
            f.write("EVAL_ERROR %r\n" % (self.get_option("eval_error")))
            f.write("MEAN_DEV %r\n" % (self.get_option("mean_dev")))
            if self.get_option("min_iterations") is not None:
                f.write("MIN_ITER %d\n" % (self.get_option("min_iterations")))
            f.write("MAX_ITER %d\n" % (self.get_option("max_iterations")))
            f.write(sd_advanced_config_section)

    def _read_solution(self,
                       symbols_filename,
                       solution_filename):
        """ Parses an SD solution file """

        # parse the symbol map
        symbol_map = {}
        with open(symbols_filename) as f:
            for line in f:
                lp_symbol, scenario_tree_id = line.strip().split()
                symbol_map[lp_symbol] = scenario_tree_id

        results = SPSolverResults()
        results.status = None
        results.solver.status = None
        results.solver.termination_condition = None
        results.solver.message = None
        xhat = {}
        try:
            with open(solution_filename, 'r') as f:
                line = f.readline()
                assert line.startswith("Problem:")
                assert line.split()[1].strip() == "pysp_model"
                line = f.readline()
                assert line.startswith("First Stage Rows:")
                line = f.readline()
                assert line.startswith("First Stage Columns:")
                line = f.readline()
                assert line.startswith("First Stage Non-zeros:")
                line = f.readline()
                assert line.startswith("Replication No.") or \
                    line.startswith("Number of replications:")
                line = f.readline()
                assert line.startswith("Status:")
                results.solver.message = line.split(":")[1].strip()
                (results.status,
                 results.solver.status,
                 results.solver.termination_condition) = \
                    _sd_status_map.get(results.solver.message,
                                       (SolutionStatus.unknown,
                                        SolverStatus.unknown,
                                        TerminationCondition.unknown))

                #
                # Objective and Bound
                #

                line = f.readline()
                assert line.startswith("Total Objective Function Upper Bound:")
                line = line.split(':')
                if line[1].strip() == '':
                    pass
                else:
                    assert len(line) == 4
                    line = line[1]
                    if "half-width" in line:
                        # we are given confidence intervals on the objective
                        line = line.split(',')
                        assert len(line) == 4
                        results.objective = float(line[0])
                        assert line[1].startswith('[')
                        assert line[2].endswith(']')
                        results.objective_interval = (float(line[1][1:]),
                                              float(line[2][:-1]))
                    else:
                        results.objective = float(line[1])
                line = f.readline()
                assert line.startswith("Total Objective Function Lower Bound:")
                line = line.split(':')
                if line[1].strip() == '':
                    pass
                else:
                    if "half-width" in line[1]:
                        # we are given confidence intervals on the bound
                        line = line[1].split(',')
                        assert len(line) == 4
                        results.bound = float(line[0])
                        assert line[1].startswith('[')
                        assert line[2].endswith(']')
                        results.bound_interval = (float(line[1][1:]),
                                          float(line[2][:-1]))
                    else:
                        results.bound = float(line[1].strip())

                #
                # Xhat
                #

                line = f.readline()
                assert line.strip() == ''
                line = f.readline()
                assert line.startswith('First Stage Solutions:')
                line = f.readline()
                assert line.startswith('   No.   Row name   Activity      Lower bound   Upper bound   Dual          Dual STDEV')
                line = f.readline()
                assert line.startswith('------ ------------ ------------- ------------- ------------- ------------- -------------')

                xhat_start_line = '   No. Column name  Activity      Lower bound   Upper bound   Reduced Cost  RC STDEV'
                line = f.readline()
                while not line.startswith(xhat_start_line):
                    line = f.readline()
                line = f.readline()
                assert line.startswith('------ ------------ ------------- ------------- ------------- ------------- -------------')
                line = f.readline().strip().split()
                while line:
                    varlabel, varvalue = line[1:3]
                    varlabel = varlabel.strip()
                    varvalue = float(varvalue)
                    xhat[symbol_map[varlabel]] = varvalue
                    line = f.readline().strip().split()

        except (IOError, OSError):
            logger.warn(
                "Exception encountered while parsing sd "
                "solution file '%s':\n%s'"
                % (solution_filename, traceback.format_exc()))
            xhat = None

        return xhat, results

def runsd_register_options(options=None):
    if options is None:
        options = PySPConfigBlock()
    SDSolver.register_options(options)
    ScenarioTreeManagerFactory.register_options(options)
    safe_register_common_option(options,
                               "verbose")
    safe_register_common_option(options,
                               "disable_gc")
    safe_register_common_option(options,
                               "profile")
    safe_register_common_option(options,
                               "traceback")
    safe_register_common_option(options,
                                "output_scenario_tree_solution")
    safe_register_common_option(options,
                                "keep_solver_files")
    safe_register_common_option(options,
                                "output_solver_log")
    safe_register_common_option(options,
                                "symbolic_solver_labels")
    return options

def runsd(options):
    """
    Construct a senario tree manager and solve it
    with the SD solver.
    """
    start_time = time.time()
    with ScenarioTreeManagerFactory(options) as sp:
        sp.initialize()
        print("")
        print("Running SD solver for stochastic "
              "programming problems")
        sd = SDSolver()
        sd_options = sd.extract_user_options_to_dict(options,
                                                     sparse=True)
        results = sd.solve(
            sp,
            options=sd_options,
            output_solver_log=options.output_solver_log,
            keep_solver_files=options.keep_solver_files,
            symbolic_solver_labels=options.symbolic_solver_labels)
        xhat = results.xhat
        del results.xhat
        print("")
        print(results)

        if options.output_scenario_tree_solution:
            print("")
            sp.scenario_tree.snapshotSolutionFromScenarios()
            sp.scenario_tree.pprintSolution()
            sp.scenario_tree.pprintCosts()

    print("")
    print("Total execution time=%.2f seconds"
          % (time.time() - start_time))
    return 0

#
# the main driver routine
#

def main(args=None):
    #
    # Top-level command that executes everything
    #

    #
    # Import plugins
    #
    import pyomo.environ

    #
    # Parse command-line options.
    #
    try:
        options = parse_command_line(
            args,
            runsd_register_options,
            prog='runsd',
            description=(
"""Optimize a stochastic program using the SD solver."""
            ))

    except SystemExit as _exc:
        # the parser throws a system exit if "-h" is specified
        # - catch it to exit gracefully.
        return _exc.code

    return launch_command(runsd,
                          options,
                          error_label="runsd: ",
                          disable_gc=options.disable_gc,
                          profile_count=options.profile,
                          traceback=options.traceback)

SPSolverFactory.register_solver("sd", SDSolver)

if __name__ == "__main__":
    sys.exit(main())



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

try:
    import matplotlib
    matplotlib.use('Agg')
except:
    pass
from pyomo.common.dependencies import (
    numpy as np, numpy_available,
    pandas as pd, pandas_available,
    scipy, scipy_available,
)
imports_present = numpy_available & pandas_available & scipy_available

import platform
is_osx = platform.mac_ver()[0] != ''

import pyutilib.th as unittest
import tempfile
import sys
import os
import shutil
import glob
import subprocess
import sys
from itertools import product

import pyomo.contrib.parmest.parmest as parmest
import pyomo.contrib.parmest.graphics as graphics
import pyomo.contrib.parmest as parmestbase
import pyomo.environ as pyo

from pyomo.opt import SolverFactory
ipopt_available = SolverFactory('ipopt').available()

if numpy_available:
    from pyomo.contrib.pynumero.asl import AmplInterface
    asl_available = AmplInterface.available()
else:
    asl_available=False

testdir = os.path.dirname(os.path.abspath(__file__))

class Object_from_string_Tester(unittest.TestCase):
    def setUp(self):
        self.instance = pyo.ConcreteModel()
        self.instance.IDX = pyo.Set(initialize=['a', 'b', 'c'])
        self.instance.x = pyo.Var(self.instance.IDX, initialize=1134)
        # TBD add a block
        if imports_present:
            np.random.seed(1134)
        
    def tearDown(self):
        pass
    
    def test_Var(self):
        # just making sure it executes
        pyo_Var_obj = parmest._object_from_string(self.instance, "x[b]")
        fixstatus = pyo_Var_obj.fixed
        self.assertEqual(fixstatus, False)

        
@unittest.skipIf(not parmest.parmest_available,
                 "Cannot test parmest: required dependencies are missing")
@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class parmest_object_Tester_RB(unittest.TestCase):
    
    def setUp(self):
        from pyomo.contrib.parmest.examples.rooney_biegler.rooney_biegler import rooney_biegler_model
           
        data = pd.DataFrame(data=[[1,8.3],[2,10.3],[3,19.0],
                                  [4,16.0],[5,15.6],[6,19.8]], columns=['hour', 'y'])
        
        theta_names = ['asymptote', 'rate_constant']
        
        def SSE(model, data):  
            expr = sum((data.y[i] - model.response_function[data.hour[i]])**2 for i in data.index)
            return expr

        solver_options = {
                'tol': 1e-8,
                }
        
        self.pest = parmest.Estimator(rooney_biegler_model, data, theta_names, SSE,
                solver_options=solver_options)

    def test_theta_est(self):
        objval, thetavals = self.pest.theta_est()
        
        self.assertAlmostEqual(objval, 4.4675, places=2)
        self.assertAlmostEqual(thetavals['asymptote'], 19.2189, places=2) # 19.1426 from the paper
        self.assertAlmostEqual(thetavals['rate_constant'], 0.5312, places=2) # 0.5311 from the paper

    @unittest.skipIf(not graphics.imports_available,
                     "parmest.graphics imports are unavailable")
    def test_bootstrap(self):
        objval, thetavals = self.pest.theta_est()
        
        num_bootstraps=10
        theta_est = self.pest.theta_est_bootstrap(num_bootstraps, return_samples=True)
        
        num_samples = theta_est['samples'].apply(len)
        self.assertTrue(len(theta_est.index), 10)
        self.assertTrue(num_samples.equals(pd.Series([6]*10)))

        del theta_est['samples']
        
        # apply cofidence region test
        CR = self.pest.confidence_region_test(theta_est, 'MVN', [0.5, 0.75, 1.0])
        
        self.assertTrue(set(CR.columns) >= set([0.5, 0.75, 1.0]))
        self.assertTrue(CR[0.5].sum() == 5) 
        self.assertTrue(CR[0.75].sum() == 7)
        self.assertTrue(CR[1.0].sum() == 10) # all true
        
        parmest.pairwise_plot(theta_est)
        parmest.pairwise_plot(theta_est, thetavals)
        parmest.pairwise_plot(theta_est, thetavals, 0.8, ['MVN', 'KDE', 'Rect'])
        
    @unittest.skipIf(not graphics.imports_available,
                     "parmest.graphics imports are unavailable")
    def test_likelihood_ratio(self):
        objval, thetavals = self.pest.theta_est()
        
        asym = np.arange(10, 30, 2)
        rate = np.arange(0, 1.5, 0.25)
        theta_vals = pd.DataFrame(list(product(asym, rate)), columns=self.pest.theta_names)
        
        obj_at_theta = self.pest.objective_at_theta(theta_vals)
        
        LR = self.pest.likelihood_ratio_test(obj_at_theta, objval, [0.8, 0.9, 1.0])
        
        self.assertTrue(set(LR.columns) >= set([0.8, 0.9, 1.0]))
        self.assertTrue(LR[0.8].sum() == 7) 
        self.assertTrue(LR[0.9].sum() == 11)
        self.assertTrue(LR[1.0].sum() == 60) # all true
        
        parmest.pairwise_plot(LR, thetavals, 0.8)

    def test_leaveNout(self):
        lNo_theta = self.pest.theta_est_leaveNout(1) 
        self.assertTrue(lNo_theta.shape == (6,2))
        
        results = self.pest.leaveNout_bootstrap_test(1, None, 3, 'Rect', [0.5, 1.0])
        self.assertTrue(len(results) == 6) # 6 lNo samples
        i = 1
        samples = results[i][0] # list of N samples that are left out
        lno_theta = results[i][1] 
        bootstrap_theta = results[i][2] 
        self.assertTrue(samples == [1]) # sample 1 was left out
        self.assertTrue(lno_theta.shape[0] == 1) # lno estimate for sample 1
        self.assertTrue(set(lno_theta.columns) >= set([0.5, 1.0]))
        self.assertTrue(lno_theta[1.0].sum() == 1) # all true
        self.assertTrue(bootstrap_theta.shape[0] == 3) # bootstrap for sample 1
        self.assertTrue(bootstrap_theta[1.0].sum() == 3) # all true
        
    def test_diagnostic_mode(self):
        self.pest.diagnostic_mode = True

        objval, thetavals = self.pest.theta_est()

        asym = np.arange(10, 30, 2)
        rate = np.arange(0, 1.5, 0.25)
        theta_vals = pd.DataFrame(list(product(asym, rate)), columns=self.pest.theta_names)
        
        obj_at_theta = self.pest.objective_at_theta(theta_vals)

        self.pest.diagnostic_mode = False

    def test_rb_main(self):
        """ test __main__ for rooney biegler """
        p = str(parmestbase.__path__)
        l = p.find("'")
        r = p.find("'", l+1)
        parmestpath = p[l+1:r]
        rbpath = parmestpath + os.sep + "examples" + os.sep + \
                   "rooney_biegler" + os.sep + "rooney_biegler.py"
        rbpath = os.path.abspath(rbpath) # paranoia strikes deep...
        if sys.version_info >= (3,5):
            ret = subprocess.run([sys.executable, rbpath])
            retcode = ret.returncode
        else:
            retcode = subprocess.call([sys.executable, rbpath])
        assert(retcode == 0)
        
    @unittest.skip("Presently having trouble with mpiexec on appveyor")
    def test_parallel_parmest(self):
        """ use mpiexec and mpi4py """
        p = str(parmestbase.__path__)
        l = p.find("'")
        r = p.find("'", l+1)
        parmestpath = p[l+1:r]
        rbpath = parmestpath + os.sep + "examples" + os.sep + \
                   "rooney_biegler" + os.sep + "rooney_biegler_parmest.py"
        rbpath = os.path.abspath(rbpath) # paranoia strikes deep...
        rlist = ["mpiexec", "--allow-run-as-root", "-n", "2", sys.executable, rbpath]
        if sys.version_info >= (3,5):
            ret = subprocess.run(rlist)
            retcode = ret.returncode
        else:
            retcode = subprocess.call(rlist)
        assert(retcode == 0)

    @unittest.skip("Most folks don't have k_aug installed")
    def test_theta_k_aug_for_Hessian(self):
        # this will fail if k_aug is not installed
        objval, thetavals, Hessian = self.pest.theta_est(solver="k_aug")
        self.assertAlmostEqual(objval, 4.4675, places=2)
        

'''
The test cases above were developed with a transcription mistake in the dataset.
This test works with the correct dataset.
'''
@unittest.skipIf(not parmest.parmest_available,
                 "Cannot test parmest: required dependencies are missing")
@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class parmest_object_Tester_RB_match_paper(unittest.TestCase):
    
    def setUp(self):
        from pyomo.contrib.parmest.examples.rooney_biegler.rooney_biegler import rooney_biegler_model
           
        data = pd.DataFrame(data=[[1,8.3],[2,10.3],[3,19.0],
                                  [4,16.0],[5,15.6],[7,19.8]], columns=['hour', 'y'])
        
        theta_names = ['asymptote', 'rate_constant']
        
        def SSE(model, data):  
            expr = sum((data.y[i] - model.response_function[data.hour[i]])**2 for i in data.index)
            return expr
        
        self.pest = parmest.Estimator(rooney_biegler_model, data, theta_names, SSE)
    
    def test_theta_est(self):
        objval, thetavals = self.pest.theta_est(calc_cov=False)
        
        self.assertAlmostEqual(objval, 4.3317112, places=2)
        self.assertAlmostEqual(thetavals['asymptote'], 19.1426, places=2) # 19.1426 from the paper
        self.assertAlmostEqual(thetavals['rate_constant'], 0.5311, places=2) # 0.5311 from the paper
    
    @unittest.skipIf(not asl_available, "Cannot test covariance matrix: required ASL dependency is missing")
    def test_theta_est_cov(self):
        objval, thetavals, cov = self.pest.theta_est(calc_cov=True)
        
        self.assertAlmostEqual(objval, 4.3317112, places=2)
        self.assertAlmostEqual(thetavals['asymptote'], 19.1426, places=2) # 19.1426 from the paper
        self.assertAlmostEqual(thetavals['rate_constant'], 0.5311, places=2) # 0.5311 from the paper
        
        # Covariance matrix
        self.assertAlmostEqual(cov[0,0], 6.30579403, places=2) # 6.22864 from paper
        self.assertAlmostEqual(cov[0,1], -0.4395341, places=2) # -0.4322 from paper
        self.assertAlmostEqual(cov[1,0], -0.4395341, places=2) # -0.4322 from paper
        self.assertAlmostEqual(cov[1,1], 0.04193591, places=2) # 0.04124 from paper
        
        ''' Why does the covariance matrix from parmest not match the paper? Parmest is
        calculating the exact reduced Hessian. The paper (Rooney and Bielger, 2001) likely
        employed the first order approximation common for nonlinear regression. The paper
        values were verified with Scipy, which uses the same first order approximation.
        The formula used in parmest was verified against equations (7-5-15) and (7-5-16) in
        "Nonlinear Parameter Estimation", Y. Bard, 1974.
        '''
        
    
        

@unittest.skipIf(not imports_present, "Cannot test parmest: required dependencies are missing")
@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class parmest_object_Tester_reactor_design(unittest.TestCase):
    
    def setUp(self):
        from pyomo.contrib.parmest.examples.reactor_design.reactor_design import reactor_design_model
          
        # Data from the design 
        data = pd.DataFrame(data=[[1.05, 10000, 3458.4, 1060.8, 1683.9, 1898.5],
                                  [1.10, 10000, 3535.1, 1064.8, 1613.3, 1893.4],
                                  [1.15, 10000, 3609.1, 1067.8, 1547.5, 1887.8],
                                  [1.20, 10000, 3680.7, 1070.0, 1486.1, 1881.6],
                                  [1.25, 10000, 3750.0, 1071.4, 1428.6, 1875.0],
                                  [1.30, 10000, 3817.1, 1072.2, 1374.6, 1868.0],
                                  [1.35, 10000, 3882.2, 1072.4, 1324.0, 1860.7],
                                  [1.40, 10000, 3945.4, 1072.1, 1276.3, 1853.1],
                                  [1.45, 10000, 4006.7, 1071.3, 1231.4, 1845.3],
                                  [1.50, 10000, 4066.4, 1070.1, 1189.0, 1837.3],
                                  [1.55, 10000, 4124.4, 1068.5, 1148.9, 1829.1],
                                  [1.60, 10000, 4180.9, 1066.5, 1111.0, 1820.8],
                                  [1.65, 10000, 4235.9, 1064.3, 1075.0, 1812.4],
                                  [1.70, 10000, 4289.5, 1061.8, 1040.9, 1803.9],
                                  [1.75, 10000, 4341.8, 1059.0, 1008.5, 1795.3],
                                  [1.80, 10000, 4392.8, 1056.0,  977.7, 1786.7],
                                  [1.85, 10000, 4442.6, 1052.8,  948.4, 1778.1],
                                  [1.90, 10000, 4491.3, 1049.4,  920.5, 1769.4],
                                  [1.95, 10000, 4538.8, 1045.8,  893.9, 1760.8]], 
                          columns=['sv', 'caf', 'ca', 'cb', 'cc', 'cd'])

        theta_names = ['k1', 'k2', 'k3']
        
        def SSE(model, data): 
            expr = (float(data['ca']) - model.ca)**2 + \
                   (float(data['cb']) - model.cb)**2 + \
                   (float(data['cc']) - model.cc)**2 + \
                   (float(data['cd']) - model.cd)**2
            return expr

        solver_options = {"max_iter": 6000}
        
        self.pest = parmest.Estimator(reactor_design_model, data,
                                      theta_names, SSE, solver_options)

    def test_theta_est(self):
        objval, thetavals = self.pest.theta_est()
        
        self.assertAlmostEqual(thetavals['k1'], 5.0/6.0, places=4) 
        self.assertAlmostEqual(thetavals['k2'], 5.0/3.0, places=4) 
        self.assertAlmostEqual(thetavals['k3'], 1.0/6000.0, places=7) 
        
@unittest.skipIf(not parmest.parmest_available,
                 "Cannot test parmest: required dependencies are missing")
@unittest.skipIf(not graphics.imports_available,
                 "parmest.graphics imports are unavailable")
@unittest.skipIf(is_osx, "Disabling graphics tests on OSX due to issue in Matplotlib, see Pyomo PR #1337")
class parmest_graphics(unittest.TestCase):
    
    def setUp(self):
        self.A = pd.DataFrame(np.random.randint(0,100,size=(100,4)), columns=list('ABCD'))
        self.B = pd.DataFrame(np.random.randint(0,100,size=(100,4)), columns=list('ABCD'))
        
    def test_pairwise_plot(self):
        parmest.pairwise_plot(self.A, alpha=0.8, distributions=['Rect', 'MVN', 'KDE'])
        
    def test_grouped_boxplot(self):
        parmest.grouped_boxplot(self.A, self.B, normalize=True, 
                                group_names=['A', 'B'])
        
    def test_grouped_violinplot(self):
        parmest.grouped_violinplot(self.A, self.B)
        
if __name__ == '__main__':
    unittest.main()



#  _________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright (c) 2014 Sandia Corporation.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  This software is distributed under the BSD License.
#  _________________________________________________________________________

import os
import time
import sys
import logging

from pyomo.opt import ProblemFormat
from pyomo.core import (Block,
                        Suffix)
from pyomo.pysp.scenariotree.manager import InvocationType
from pyomo.pysp.util.config import (PySPConfigValue,
                                    PySPConfigBlock,
                                    safe_register_common_option,
                                    safe_register_unique_option,
                                    _domain_must_be_str)
from pyomo.pysp.scenariotree.manager import \
    (ScenarioTreeManagerClientSerial,
     ScenarioTreeManagerClientPyro)
from pyomo.pysp.scenariotree.util import \
    scenario_tree_id_to_pint32
from pyomo.pysp.util.misc import (parse_command_line,
                                  launch_command)

thisfile = os.path.abspath(__file__)

logger = logging.getLogger('pyomo.pysp')

_objective_weight_suffix_name = "schurip_objective_weight"
_variable_id_suffix_name = "schurip_variable_id"

def _write_bundle_nl(worker,
                     bundle,
                     output_directory,
                     io_options):

    assert os.path.exists(output_directory)

    bundle_instance = worker._bundle_binding_instance_map[bundle.name]
    assert not hasattr(bundle_instance, ".schuripopt")
    tmpblock = Block(concrete=True)
    bundle_instance.add_component(".schuripopt", tmpblock)

    #
    # linking variable suffix
    #
    tmpblock.add_component(_variable_id_suffix_name,
                           Suffix(direction=Suffix.EXPORT,
                                  datatype=Suffix.INT))
    linking_suffix = getattr(tmpblock, _variable_id_suffix_name)

    # Loop over all nodes for the bundle except the leaf nodes,
    # which have no blended variables
    scenario_tree = worker.scenario_tree
    for stage in bundle.scenario_tree.stages[:-1]:
        for _node in stage.nodes:
            # get the node of off the real scenario tree
            # as this has the linked variable information
            node = scenario_tree.get_node(_node.name)
            node_name = node.name
            master_variable = bundle_instance.find_component(
                "MASTER_BLEND_VAR_"+str(node.name))
            for variable_id in node._standard_variable_ids:
                # Assumes ASL uses 4-byte, signed integers to store suffixes,
                # and we need positive suffix values
                linking_suffix[master_variable[variable_id]] = \
                    scenario_tree_id_to_pint32(node_name, variable_id)
    # make sure the conversion from scenario tree id to int
    # did not have any collisions
    _ids = list(linking_suffix.values())
    assert len(_ids) == len(set(_ids))

    #
    # objective weight suffix
    #
    tmpblock.add_component(_objective_weight_suffix_name,
                           Suffix(direction=Suffix.EXPORT))
    getattr(tmpblock, _objective_weight_suffix_name)[bundle_instance] = \
        bundle.probability

    # take care to disable any advanced preprocessing flags since we
    # are not going through the scenario tree manager solver interface
    # TODO: resolve this preprocessing mess
    block_attrs = []
    for block in bundle_instance.block_data_objects(active=True):
        attrs = []
        for attr_name in ("_gen_obj_repn",
                          "_gen_con_repn"):
            if hasattr(block, attr_name):
                attrs.append((attr_name, getattr(block, attr_name)))
                setattr(block, attr_name, True)
        if len(attrs):
            block_attrs.append((block, attrs))

    output_filename = os.path.join(output_directory,
                                   str(bundle.name)+".nl")
    # write the model and obtain the symbol_map
    _, smap_id = bundle_instance.write(
        output_filename,
        format=ProblemFormat.nl,
        io_options=io_options)
    symbol_map = bundle_instance.solutions.symbol_map[smap_id]

    # reset preprocessing flags
    # TODO: resolve this preprocessing mess
    for block, attrs in block_attrs:
        for attr_name, attr_val in attrs:
            setattr(block, attr_name, attr_val)

    bundle_instance.del_component(tmpblock)

    return output_filename, symbol_map

def _write_scenario_nl(worker,
                       scenario,
                       output_directory,
                       io_options):

    assert os.path.exists(output_directory)
    instance = scenario._instance
    assert not hasattr(instance, ".schuripopt")
    tmpblock = Block(concrete=True)
    instance.add_component(".schuripopt", tmpblock)

    #
    # linking variable suffix
    #
    bySymbol = instance._ScenarioTreeSymbolMap.bySymbol
    tmpblock.add_component(_variable_id_suffix_name,
                           Suffix(direction=Suffix.EXPORT,
                                  datatype=Suffix.INT))
    linking_suffix = getattr(tmpblock, _variable_id_suffix_name)

    # Loop over all nodes for the scenario except the leaf node,
    # which has no blended variables
    for node in scenario._node_list[:-1]:
        node_name = node.name
        for variable_id in node._standard_variable_ids:
            # Assumes ASL uses 4-byte, signed integers to store suffixes,
            # and we need positive suffix values
            linking_suffix[bySymbol[variable_id]] = \
                scenario_tree_id_to_pint32(node_name, variable_id)
    # make sure the conversion from scenario tree id to int
    # did not have any collisions
    _ids = list(linking_suffix.values())
    assert len(_ids) == len(set(_ids))

    #
    # objective weight suffix
    #
    tmpblock.add_component(_objective_weight_suffix_name,
                           Suffix(direction=Suffix.EXPORT))
    getattr(tmpblock, _objective_weight_suffix_name)[instance] = \
        scenario.probability

    # take care to disable any advanced preprocessing flags since we
    # are not going through the scenario tree manager solver interface
    # TODO: resolve this preprocessing mess
    block_attrs = []
    for block in instance.block_data_objects(active=True):
        attrs = []
        for attr_name in ("_gen_obj_repn",
                          "_gen_con_repn"):
            if hasattr(block, attr_name):
                attrs.append((attr_name, getattr(block, attr_name)))
                setattr(block, attr_name, True)
        if len(attrs):
            block_attrs.append((block, attrs))

    output_filename = os.path.join(output_directory,
                                   str(scenario.name)+".nl")

    # write the model and obtain the symbol_map
    _, smap_id = instance.write(
        output_filename,
        format=ProblemFormat.nl,
        io_options=io_options)
    symbol_map = instance.solutions.symbol_map[smap_id]

    # reset preprocessing flags
    # TODO: resolve this preprocessing mess
    for block, attrs in block_attrs:
        for attr_name, attr_val in attrs:
            setattr(block, attr_name, attr_val)

    instance.del_component(tmpblock)

    return output_filename, symbol_map

def _write_problem_list_file(scenario_tree,
                             problem_list_filename,
                             ignore_bundles=False):

    #
    # Write list of subproblems to file
    #
    subproblem_type = None
    with open(problem_list_filename, 'w') as f:
        if (not ignore_bundles) and scenario_tree.contains_bundles():
            subproblem_type = "bundles"
            for bundle in scenario_tree.bundles:
                f.write(str(bundle.name)+".nl")
                f.write("\n")
        else:
            subproblem_type = "scenarios"
            for scenario in scenario_tree.scenarios:
                f.write(str(scenario.name)+".nl")
                f.write("\n")
    assert subproblem_type is not None
    return subproblem_type

def EXTERNAL_write_nl(worker,
                      working_directory,
                      subproblem_type,
                      io_options):

    if subproblem_type == 'bundles':
        assert worker.scenario_tree.contains_bundles()
        for bundle in worker.scenario_tree.bundles:
            fname, symbol_map = _write_bundle_nl(
                worker,
                bundle,
                working_directory,
                io_options)
    else:
        assert subproblem_type == 'scenarios'
        orig_parents = {}
        if worker.scenario_tree.contains_bundles():
            for scenario in worker.scenario_tree.scenarios:
                if scenario._instance._parent is not None:
                    orig_parents[scenario] = scenario._instance._parent
                    scenario._instance._parent = None
                    assert not scenario._instance_objective.active
                    scenario._instance_objective.activate()
        try:
            for scenario in worker.scenario_tree.scenarios:
                fname, symbol_map = _write_scenario_nl(
                    worker,
                    scenario,
                    working_directory,
                    io_options)
        finally:
            for scenario, parent in orig_parents.items():
                scenario._instance._parent = parent
                assert scenario._instance_objective.active
                scenario._instance_objective.deactivate()

def write_schuripopt_files(manager,
                           output_directory,
                           ignore_bundles=False,
                           io_options=None,
                           verbose=False):

    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    problem_list_filename = os.path.join(output_directory,
                                         "PySP_Subproblems.txt")

    scenario_tree = manager.scenario_tree

    subproblem_type = _write_problem_list_file(
        scenario_tree,
        problem_list_filename,
        ignore_bundles=ignore_bundles)

    assert subproblem_type in ('scenarios','bundles')

    manager.invoke_function(
        "EXTERNAL_write_nl",
        thisfile,
        invocation_type=InvocationType.Single,
        function_args=(output_directory,
                       subproblem_type,
                       io_options))

def convertschuripopt_register_options(options=None):
    if options is None:
        options = PySPConfigBlock()
    safe_register_common_option(options, "disable_gc")
    safe_register_common_option(options, "profile")
    safe_register_common_option(options, "traceback")
    safe_register_common_option(options, "verbose")
    safe_register_common_option(options, "symbolic_solver_labels")
    safe_register_unique_option(
        options,
        "output_directory",
        PySPConfigValue(
            ".",
            domain=_domain_must_be_str,
            description=(
                "The directory in which all SchurIpopt files "
                "will be stored. Default is '.'."
            ),
            doc=None,
            visibility=0))
    safe_register_unique_option(
        options,
        "ignore_bundles",
        PySPConfigValue(
            False,
            domain=bool,
            description=(
                "Ignore bundles when converting the SP to "
                "SchurIpopt input files."
            ),
            doc=None,
            visibility=0))
    safe_register_common_option(options, "scenario_tree_manager")
    ScenarioTreeManagerClientSerial.register_options(options)
    ScenarioTreeManagerClientPyro.register_options(options)

    return options

#
# Convert a PySP scenario tree formulation to DDSIP input files
#

def convertschuripopt(options):
    """
    Construct a senario tree manager and write the
    schuripopt input files.
    """
    import pyomo.environ

    start_time = time.time()

    io_options = {'symbolic_solver_labels':
                  options.symbolic_solver_labels}

    assert not options.compile_scenario_instances

    manager_class = None
    if options.scenario_tree_manager == 'serial':
        manager_class = ScenarioTreeManagerClientSerial
    elif options.scenario_tree_manager == 'pyro':
        manager_class = ScenarioTreeManagerClientPyro

    with manager_class(options) as scenario_tree_manager:
        scenario_tree_manager.initialize()
        files = write_schuripopt_files(
            scenario_tree_manager,
            options.output_directory,
            ignore_bundles=options.ignore_bundles,
            io_options=io_options,
            verbose=options.verbose)

    end_time = time.time()

    print("SchurIpopt files written to directory: %s"
          % (options.output_directory))
    print("")
    print("Total execution time=%.2f seconds"
          % (end_time - start_time))

#
# the main driver routine
#

def main(args=None):
    #
    # Top-level command that executes everything
    #

    #
    # Import plugins
    #
    import pyomo.environ

    #
    # Parse command-line options.
    #
    try:
        options = parse_command_line(
            args,
            convertschuripopt_register_options,
            prog='convertschuripopt',
            description=(
"""Optimize a stochastic program using the SchurIpopt solver."""
            ))

    except SystemExit as _exc:
        # the parser throws a system exit if "-h" is specified
        # - catch it to exit gracefully.
        return _exc.code

    return launch_command(convertschuripopt,
                          options,
                          error_label="convertschuripopt: ",
                          disable_gc=options.disable_gc,
                          profile_count=options.profile,
                          traceback=options.traceback)

if __name__ == "__main__":
    sys.exit(main())



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import json
import sys
import os

from pyomo.pysp.tests.examples.ph_checker import \
    _update_exception_message, \
    assert_value_equals, \
    assert_float_equals, \
    validate_probabilities, \
    validate_variable_set, \
    validate_leaf_stage_costs

def validate_nonanticipativity(options,
                               scenario_tree,
                               scenario_solutions,
                               node_solutions):

    # Check that variables with non-anticipativity constraints have
    # converged
    for node_name, node in scenario_tree['nodes'].items():
        assert node_name == node['name']
        node_sol = node_solutions[node_name]
        for scenario_name in node['scenarios']:
            scenario_sol = scenario_solutions[scenario_name]
            # variables, objective, stage costs
            for variable_name, variable_node_sol in node_sol['variables'].items():
                variable_scenario_sol = scenario_sol['variables'][variable_name]
                try:
                    assert_float_equals(variable_node_sol['solution'],
                                        variable_scenario_sol['value'],
                                        options.absolute_tolerance)
                except AssertionError as e:
                    msg = ""
                    msg += ("Problem validating nonanticipativity within "
                            "node %s for variable %s in scenario %s. "
                            "Variable value in scenario solution does not "
                            "match node solution.\n"
                            % (node_name, variable_name, scenario_name))
                    msg += ("Node value: %r\n"
                            % (variable_node_sol['solution']))
                    msg += ("Scenario value: %r\n"
                            % (variable_scenario_sol['value']))
                    _update_exception_message(msg, e)
                    raise
                try:
                    assert_value_equals(variable_node_sol['fixed'],
                                        variable_scenario_sol['fixed'])
                except AssertionError as e:
                    msg = ""
                    msg += ("Problem validating nonanticipativity within "
                            "node %s for variable %s in scenario %s. "
                            "Variable fixed status in scenario solution does not "
                            "match node fixed status.\n"
                            % (node_name, variable_name, scenario_name))
                    msg += ("Node fixed status: %s\n"
                            % (variable_node_sol['fixed']))
                    msg += ("Scenario fixed status: %s\n"
                            % (variable_scenario_sol['fixed']))
                    _update_exception_message(msg, e)
                    raise

def validate_expected_node_costs(options,
                                 scenario_tree,
                                 scenario_solutions,
                                 node_solutions):

    # Check that reported expected node costs make sense
    # and that the expected node cost of the root node
    # equals the ef objective function

    # sorted by time
    sorted_stages = \
        sorted(scenario_tree['stages'].values(),key=lambda x: x['order'])
    # leaf to root
    reverse_sorted_stages = reversed(sorted_stages)
    node_expected_costs = \
        dict((node_name,0.0) for node_name in scenario_tree['nodes'])
    for stage in reverse_sorted_stages:
        assert stage is scenario_tree['stages'][stage['name']]
        stage_name = stage['name']
        for node_name in stage['nodes']:
            node = scenario_tree['nodes'][node_name]
            assert node['name'] == node_name
            node_scenario_names = node['scenarios']
            
            single_scenario_name = node_scenario_names[0]
            single_scenario_stage_cost = \
                scenario_solutions[single_scenario_name]\
                ['stage costs'][stage_name]
            node_cost = 0.0
            for scenario_name in node_scenario_names:
                scenario = scenario_tree['scenarios'][scenario_name]
                scenario_stage_cost = \
                    scenario_solutions[scenario_name]['stage costs'][stage_name]
                node_cost += \
                    scenario['probability'] * \
                    scenario_solutions[scenario_name]['stage costs'][stage_name]

                try:
                    assert_float_equals(single_scenario_stage_cost,
                                        scenario_stage_cost,
                                        options.absolute_tolerance)
                except AssertionError as e:
                    msg = ""
                    msg += ("Problem validating stage costs "
                            "for stage %s of scenarios within node %s. "
                            "Not all scenarios report the same stage cost.\n"
                            % (stage_name, node_name))
                    msg += ("Scenario %s stage cost: %r\n"
                            % (scenario_name, scenario_stage_cost))
                    msg += ("Scenario %s stage cost: %r\n"
                            % (single_scenario_name, single_scenario_stage_cost))
                    _update_exception_message(msg, e)
                    raise

            node_cost /= node['probability']
            node_expected_costs[node_name] += node_cost
            for child_node_name in node['children']:
                child_node = scenario_tree['nodes'][child_node_name]
                node_expected_costs[node_name] += \
                    child_node['conditional probability'] * \
                    node_expected_costs[child_node_name]

                del child_node

            del node
            del node_scenario_names
            del node_cost
            del single_scenario_stage_cost

        del stage_name

    for node_name, expected_cost in node_expected_costs.items():
        node_sol = node_solutions[node_name]
        try:
            assert_float_equals(expected_cost,
                                node_sol['expected cost'],
                                options.absolute_tolerance)
        except AssertionError as e:
            msg = ""
            msg += ("Problem validating node expected cost for node %s\n"
                    % (node_name))
            msg += ("Node expected cost in solution: %r\n"
                    % (node_sol['expected cost']))
            msg += ("Computed average from stage costs: %r\n"
                    % (expected_cost))
            _update_exception_message(msg, e)
            raise

        del node_sol

    # get the root node
    assert len(scenario_tree['stages'][sorted_stages[0]['name']]['nodes']) == 1
    root_node_name = scenario_tree['stages'][sorted_stages[0]['name']]['nodes'][0]
    root_node_expected_cost = node_expected_costs[root_node_name]
    expected_scenario_objective = 0.0
    expected_scenario_cost1 = 0.0
    expected_scenario_cost2 = 0.0
    for scenario_name, scenario in scenario_tree['scenarios'].items():
        scenario_sol = scenario_solutions[scenario_name]
        expected_scenario_objective += \
            scenario['probability']*scenario_sol['objective']
        expected_scenario_cost1 += \
            scenario['probability']*scenario_sol['cost']
        expected_scenario_cost2 += \
            scenario['probability'] * \
            sum(stage_cost for stage_cost in scenario_sol['stage costs'].values())


    try:
        assert_float_equals(root_node_expected_cost,
                            expected_scenario_cost1,
                            options.absolute_tolerance)
    except AssertionError as e:
        msg = ""
        msg += ("Problem validating root node expected cost\n")
        msg += ("Root node expected cost in solution: %r\n"
                % (root_node_expected_cost))
        msg += ("Computed average from scenario costs in solution: %r\n"
                % (expected_scenario_cost1))
        _update_exception_message(msg, e)
        raise

    try:
        assert_float_equals(root_node_expected_cost,
                            expected_scenario_cost2,
                            options.absolute_tolerance)
    except AssertionError as e:
        msg = ""
        msg += ("Problem validating root node expected cost\n")
        msg += ("Root node expected cost in solution: %r\n"
                % (root_node_expected_cost))
        msg += ("Computed average from scenario stage costs in solution: %r\n"
                % (expected_scenario_cost2))
        _update_exception_message(msg, e)
        raise

    try:
        assert_float_equals(root_node_expected_cost,
                            expected_scenario_objective,
                            options.absolute_tolerance)
    except AssertionError as e:
        msg = ""
        msg += ("Problem validating root node expected cost\n")
        msg += ("Root node expected cost in solution: %r\n"
                % (root_node_expected_cost))
        msg += ("Computed average from scenario objectives in solution: %r\n"
                % (expected_scenario_objective))
        _update_exception_message(msg, e)
        raise

def validate_ef(options):

    results = None
    with open(options.ef_solution_filename) as f:
        results = json.load(f)

    scenario_tree = results['scenario tree']

    try:
        validate_probabilities(options,
                               scenario_tree)
    except AssertionError as e:
        msg = "FAILURE:\n"
        msg += ("Problem validating probabilities for scenario "
                "tree in ef solution from file: %s"
                % (options.ef_solution_filename))
        _update_exception_message(msg, e)
        raise

    scenario_solutions = results['scenario solutions']
    node_solutions = results['node solutions']

    assert set(scenario_tree['nodes'].keys()) \
        == set(node_solutions.keys())
    assert set(scenario_tree['scenarios']) \
        == set(scenario_solutions.keys())

    try:
        validate_variable_set(options,
                              scenario_tree,
                              scenario_solutions,
                              node_solutions)
    except AssertionError as e:
        msg = "FAILURE:\n"
        msg += ("Problem validating variable set "
                "in ef solution from file: %s"
                % (options.ef_solution_filename))
        _update_exception_message(msg, e)
        raise

    try:
        validate_nonanticipativity(options,
                                   scenario_tree,
                                   scenario_solutions,
                                   node_solutions)
    except AssertionError as e:
        msg = "FAILURE:\n"
        msg += ("Problem validating nonanticipativity "
                "in ef solution from file: %s"
                % (options.ef_solution_filename))
        _update_exception_message(msg, e)
        raise

    try:
        validate_leaf_stage_costs(options,
                                  scenario_tree,
                                  scenario_solutions,
                                  node_solutions)
    except AssertionError as e:
        msg = "FAILURE:\n"
        msg += ("Problem validating leaf-stage costs "
                "in ef solution from file: %s"
                % (options.ef_solution_filename))
        _update_exception_message(msg, e)
        raise

    try:
        validate_expected_node_costs(options,
                                     scenario_tree,
                                     scenario_solutions,
                                     node_solutions)
    except AssertionError as e:
        msg = "FAILURE:\n"
        msg += ("Problem validating expected node costs "
                "in ef solution from file: %s"
                % (options.ef_solution_filename))
        _update_exception_message(msg, e)
        raise

def construct_parser():

    import argparse
    parser = argparse.ArgumentParser(
        description=('Validate a JSON ef solution output file that '
                     'was created the jsonsolutionwriter plugin'))
    parser.add_argument('ef_solution_filename', metavar='JSON_SOLUTION', type=str,
                        action='store', default=None,
                        help=('JSON formatted output file to '
                              'validate'))
    parser.add_argument('-t','--tolerance',dest='absolute_tolerance',
                        action='store', default=1e-6, type=float,
                        help=('absolute tolerance used when comparing '
                              'floating point values (default: %(default)s)'))

    return parser


def main(args=None):

    parser = construct_parser()
    options = parser.parse_args(args=args)
    assert os.path.exists(options.ef_solution_filename)
    validate_ef(options)

if __name__ == "__main__":

    import sys
    main(args=sys.argv[1:])



import sys
import logging
from pyutilib.misc.timing import TicTocTimer

_logger = logging.getLogger('pyomo.common.timing')
_logger.propagate = False
_logger.setLevel(logging.WARNING)

def report_timing(stream=True):
    if stream:
        _logger.setLevel(logging.INFO)
        if stream is True:
            stream = sys.stdout
        handler = logging.StreamHandler(stream)
        handler.setFormatter(logging.Formatter("      %(message)s"))
        _logger.addHandler(handler)
        return handler
    else:
        _logger.setLevel(logging.WARNING)
        for h in _logger.handlers:
            _logger.removeHandler(h)

_construction_logger = logging.getLogger('pyomo.common.timing.construction')
class ConstructionTimer(object):
    fmt = "%%6.%df seconds to construct %s %s; %d %s total"
    def __init__(self, obj):
        self.obj = obj
        self.timer = TicTocTimer()

    def report(self):
        # Record the elapsed time, as some log handlers may not
        # immediately generate the messge string
        self.timer = self.timer.toc(msg="")
        _construction_logger.info(self)

    def __str__(self):
        total_time = self.timer
        try:
            idx = len(self.obj.index_set())
        except AttributeError:
            idx = 1
        try:
            name = self.obj.name
        except RuntimeError:
            try:
                name = self.obj.local_name
            except RuntimeError:
                name = '(unknown)'
        except AttributeError:
            name = '(unknown)'
        try:
            _type = self.obj.ctype.__name__
        except AttributeError:
            _type = type(self.obj).__name__
        try:
            return self.fmt % ( 2 if total_time>=0.005 else 0,
                                _type,
                                name,
                                idx,
                                'indicies' if idx > 1 else 'index',
                            ) % total_time
        except TypeError:
            return "ConstructionTimer object for %s %s; %s elapsed seconds" % (
                _type,
                name,
                self.timer.toc("") )


_transform_logger = logging.getLogger('pyomo.common.timing.transformation')
class TransformationTimer(object):
    fmt = "%%6.%df seconds to apply Transformation %s%s"
    def __init__(self, obj, mode=None):
        self.obj = obj
        if mode is None:
            self.mode = ''
        else:
            self.mode = " (%s)" % (mode,)
        self.timer = TicTocTimer()

    def report(self):
        # Record the elapsed time, as some log handlers may not
        # immediately generate the message string
        self.timer = self.timer.toc(msg="")
        _transform_logger.info(self)

    def __str__(self):
        total_time = self.timer
        name = self.obj.__class__.__name__
        try:
            return self.fmt % ( 2 if total_time>=0.005 else 0,
                                name,
                                self.mode,
                            ) % total_time
        except TypeError:
            return "TransformationTimer object for %s; %s elapsed seconds" % (
                name,
                self.timer.toc("") )



#
# a model that declares the wrong component type
# within an annotation
#
from pyomo.pysp.tests.convert.utils import *

pysp_scenario_tree_model_callback = \
    simple_twostage_scenario_tree

def pysp_instance_creation_callback(scenario_name, node_names):
    model = simple_twostage_model()
    assert model.find_component("b") is None
    model.b = Block()
    model.b.c = Constraint(expr=model.y == 1)
    model.smat = StochasticConstraintBodyAnnotation()
    model.smat.declare(model.o)
    return model
